@@ -170,14 +170,14 @@ func testOutput(t *testing.T, dlvbin, output string, delveCmds []string) (stdout
 
 	_, err = os.Stat(debugbin)
 	if err == nil {
-		if strings.ToLower(os.Getenv("TRAVIS")) == "true" && runtime.GOOS == "windows" {
-			// Sometimes delve on Travis on Windows can't remove the built binary before
-			// exiting and gets an "Access is denied" error when trying.
-			// Just ignore it.
-			// See: https://travis-ci.com/go-delve/delve/jobs/296325131
-			return
+		// Sometimes delve on Windows can't remove the built binary before
+		// exiting and gets an "Access is denied" error when trying.
+		// See: https://travis-ci.com/go-delve/delve/jobs/296325131)
+		// We have added a delay to gobuild.Remove, but to avoid any test
+		// flakiness, we guard against this failure here as well.
+		if runtime.GOOS != "windows" || !strings.Contains(err.Error(), "Access is denied") {
+			t.Errorf("running %q: file %v was not deleted\nstdout is %q, stderr is %q", delveCmds, debugbin, stdout, stderr)
 		}
-		t.Errorf("running %q: file %v was not deleted\nstdout is %q, stderr is %q", delveCmds, debugbin, stdout, stderr)
 		return
 	}
 	if !os.IsNotExist(err) {
@@ -322,21 +322,24 @@ func TestRedirect(t *testing.T) {
 
 	// and detach from and kill the headless instance
 	client := rpc2.NewClient(listenAddr)
-	if err := client.Detach(true); err != nil {
-		t.Fatalf("error detaching from headless instance: %v", err)
-	}
+	_ = client.Detach(true)
 	cmd.Wait()
 }
 
 func checkAutogenDoc(t *testing.T, filename, gencommand string, generated []byte) {
 	saved := slurpFile(t, filepath.Join(projectRoot(), filename))
 
+	saved = bytes.ReplaceAll(saved, []byte("\r\n"), []byte{'\n'})
+	generated = bytes.ReplaceAll(generated, []byte("\r\n"), []byte{'\n'})
+
 	if len(saved) != len(generated) {
+		t.Logf("generated %q saved %q\n", generated, saved)
 		t.Fatalf("%s: needs to be regenerated; run %s", filename, gencommand)
 	}
 
 	for i := range saved {
 		if saved[i] != generated[i] {
+			t.Logf("generated %q saved %q\n", generated, saved)
 			t.Fatalf("%s: needs to be regenerated; run %s", filename, gencommand)
 		}
 	}
