@@ -3,12 +3,13 @@ package line
 import (
 	"bytes"
 	"encoding/binary"
-	"path/filepath"
+	"path"
 	"strings"
 
 	"github.com/go-delve/delve/pkg/dwarf/util"
 )
 
+// DebugLinePrologue prologue of .debug_line data.
 type DebugLinePrologue struct {
 	UnitLength     uint32
 	Version        uint16
@@ -22,6 +23,7 @@ type DebugLinePrologue struct {
 	StdOpLengths   []uint8
 }
 
+// DebugLineInfo info of .debug_line data.
 type DebugLineInfo struct {
 	Prologue     *DebugLinePrologue
 	IncludeDirs  []string
@@ -46,6 +48,7 @@ type DebugLineInfo struct {
 	endSeqIsValid      bool
 }
 
+// FileEntry file entry in File Name Table.
 type FileEntry struct {
 	Path        string
 	DirIdx      uint64
@@ -86,11 +89,19 @@ func Parse(compdir string, buf *bytes.Buffer, logfn func(string, ...interface{})
 
 	parseDebugLinePrologue(dbl, buf)
 	if dbl.Prologue.Version >= 5 {
-		parseIncludeDirs5(dbl, buf)
-		parseFileEntries5(dbl, buf)
+		if !parseIncludeDirs5(dbl, buf) {
+			return nil
+		}
+		if !parseFileEntries5(dbl, buf) {
+			return nil
+		}
 	} else {
-		parseIncludeDirs2(dbl, buf)
-		parseFileEntries2(dbl, buf)
+		if !parseIncludeDirs2(dbl, buf) {
+			return nil
+		}
+		if !parseFileEntries2(dbl, buf) {
+			return nil
+		}
 	}
 
 	// Instructions size calculation breakdown:
@@ -132,20 +143,30 @@ func parseDebugLinePrologue(dbl *DebugLineInfo, buf *bytes.Buffer) {
 }
 
 // parseIncludeDirs2 parses the directory table for DWARF version 2 through 4.
-func parseIncludeDirs2(info *DebugLineInfo, buf *bytes.Buffer) {
+func parseIncludeDirs2(info *DebugLineInfo, buf *bytes.Buffer) bool {
 	for {
-		str, _ := util.ParseString(buf)
+		str, err := util.ParseString(buf)
+		if err != nil {
+			if info.Logf != nil {
+				info.Logf("error reading string: %v", err)
+			}
+			return false
+		}
 		if str == "" {
 			break
 		}
 
 		info.IncludeDirs = append(info.IncludeDirs, str)
 	}
+	return true
 }
 
 // parseIncludeDirs5 parses the directory table for DWARF version 5.
-func parseIncludeDirs5(info *DebugLineInfo, buf *bytes.Buffer) {
+func parseIncludeDirs5(info *DebugLineInfo, buf *bytes.Buffer) bool {
 	dirEntryFormReader := readEntryFormat(buf, info.Logf)
+	if dirEntryFormReader == nil {
+		return false
+	}
 	dirCount, _ := util.DecodeULEB128(buf)
 	info.IncludeDirs = make([]string, 0, dirCount)
 	for i := uint64(0); i < dirCount; i++ {
@@ -165,13 +186,23 @@ func parseIncludeDirs5(info *DebugLineInfo, buf *bytes.Buffer) {
 			case _DW_LNCT_MD5:
 			}
 		}
+		if dirEntryFormReader.err != nil {
+			if info.Logf != nil {
+				info.Logf("error reading directory entries table: %v", dirEntryFormReader.err)
+			}
+			return false
+		}
 	}
+	return true
 }
 
 // parseFileEntries2 parses the file table for DWARF 2 through 4
-func parseFileEntries2(info *DebugLineInfo, buf *bytes.Buffer) {
+func parseFileEntries2(info *DebugLineInfo, buf *bytes.Buffer) bool {
 	for {
 		entry := readFileEntry(info, buf, true)
+		if entry == nil {
+			return false
+		}
 		if entry.Path == "" {
 			break
 		}
@@ -179,48 +210,75 @@ func parseFileEntries2(info *DebugLineInfo, buf *bytes.Buffer) {
 		info.FileNames = append(info.FileNames, entry)
 		info.Lookup[entry.Path] = entry
 	}
+	return true
 }
 
 func readFileEntry(info *DebugLineInfo, buf *bytes.Buffer, exitOnEmptyPath bool) *FileEntry {
 	entry := new(FileEntry)
 
-	entry.Path, _ = util.ParseString(buf)
+	var err error
+	entry.Path, err = util.ParseString(buf)
+	if err != nil {
+		if info.Logf != nil {
+			info.Logf("error reading file entry: %v", err)
+		}
+		return nil
+	}
 	if entry.Path == "" && exitOnEmptyPath {
 		return entry
 	}
 
 	if info.normalizeBackslash {
-		entry.Path = strings.Replace(entry.Path, "\\", "/", -1)
+		entry.Path = strings.ReplaceAll(entry.Path, "\\", "/")
 	}
 
 	entry.DirIdx, _ = util.DecodeULEB128(buf)
 	entry.LastModTime, _ = util.DecodeULEB128(buf)
 	entry.Length, _ = util.DecodeULEB128(buf)
-	if !filepath.IsAbs(entry.Path) {
+	if !pathIsAbs(entry.Path) {
 		if entry.DirIdx >= 0 && entry.DirIdx < uint64(len(info.IncludeDirs)) {
-			entry.Path = filepath.Join(info.IncludeDirs[entry.DirIdx], entry.Path)
+			entry.Path = path.Join(info.IncludeDirs[entry.DirIdx], entry.Path)
 		}
 	}
 
 	return entry
 }
 
+// pathIsAbs returns true if this is an absolute path.
+// We can not use path.IsAbs because it will not recognize windows paths as
+// absolute. We also can not use filepath.Abs because we want this
+// processing to be independent of the host operating system (we could be
+// reading an executable file produced on windows on a unix machine or vice
+// versa).
+func pathIsAbs(s string) bool {
+	if len(s) >= 1 && s[0] == '/' {
+		return true
+	}
+	if len(s) >= 2 && s[1] == ':' && (('a' <= s[0] && s[0] <= 'z') || ('A' <= s[0] && s[0] <= 'Z')) {
+		return true
+	}
+	return false
+}
+
 // parseFileEntries5 parses the file table for DWARF 5
-func parseFileEntries5(info *DebugLineInfo, buf *bytes.Buffer) {
+func parseFileEntries5(info *DebugLineInfo, buf *bytes.Buffer) bool {
 	fileEntryFormReader := readEntryFormat(buf, info.Logf)
+	if fileEntryFormReader == nil {
+		return false
+	}
 	fileCount, _ := util.DecodeULEB128(buf)
 	info.FileNames = make([]*FileEntry, 0, fileCount)
 	for i := 0; i < int(fileCount); i++ {
 		fileEntryFormReader.reset()
 		for fileEntryFormReader.next(buf) {
 			entry := new(FileEntry)
-			var path string
+			var p string
 			var diridx int = -1
 
 			switch fileEntryFormReader.contentType {
 			case _DW_LNCT_path:
 				if fileEntryFormReader.formCode != _DW_FORM_string {
-					path = fileEntryFormReader.str
+					p = fileEntryFormReader.str
 				} else {
 					//TODO(aarzilli): support debug_string, debug_line_str
 					info.Logf("unsupported string form %#x", fileEntryFormReader.formCode)
@@ -235,12 +293,23 @@ func parseFileEntries5(info *DebugLineInfo, buf *bytes.Buffer) {
 				// not implemented
 			}
 
-			if diridx >= 0 && !filepath.IsAbs(path) && diridx < len(info.IncludeDirs) {
-				path = filepath.Join(info.IncludeDirs[diridx], path)
+			if info.normalizeBackslash {
+				p = strings.ReplaceAll(p, "\\", "/")
 			}
-			entry.Path = path
+
+			if diridx >= 0 && !pathIsAbs(p) && diridx < len(info.IncludeDirs) {
+				p = path.Join(info.IncludeDirs[diridx], p)
+			}
+			entry.Path = p
 			info.FileNames = append(info.FileNames, entry)
 			info.Lookup[entry.Path] = entry
 		}
+		if fileEntryFormReader.err != nil {
+			if info.Logf != nil {
+				info.Logf("error reading file entries table: %v", fileEntryFormReader.err)
+			}
+			return false
+		}
 	}
+	return true
 }
