@@ -69,9 +69,6 @@ type BinaryInfo struct {
 
 	lastModified time.Time // Time the executable of this process was last modified
 
-	closer         io.Closer
-	sepDebugCloser io.Closer
-
 	// PackageMap maps package names to package paths, needed to lookup types inside DWARF info.
 	// On Go1.12 this mapping is determined by using the last element of a package path, for example:
 	//   github.com/go-delve/delve
@@ -123,12 +120,13 @@ var (
 		elf.EM_386:     true,
 	}
 
-	supportedWindowsArch = map[PEMachine]bool{
-		IMAGE_FILE_MACHINE_AMD64: true,
+	supportedWindowsArch = map[_PEMachine]bool{
+		_IMAGE_FILE_MACHINE_AMD64: true,
 	}
 
 	supportedDarwinArch = map[macho.Cpu]bool{
 		macho.CpuAmd64: true,
+		macho.CpuArm64: true,
 	}
 )
 
@@ -226,19 +224,19 @@ func FirstPCAfterPrologue(p Process, fn *Function, sameline bool) (uint64, error
 	return pc, nil
 }
 
-// CpuArch is a stringer interface representing CPU architectures.
-type CpuArch interface {
+// cpuArch is a stringer interface representing CPU architectures.
+type cpuArch interface {
 	String() string
 }
 
 // ErrUnsupportedArch is returned when attempting to debug a binary compiled for an unsupported architecture.
 type ErrUnsupportedArch struct {
 	os      string
-	cpuArch CpuArch
+	cpuArch cpuArch
 }
 
 func (e *ErrUnsupportedArch) Error() string {
-	var supportArchs []CpuArch
+	var supportArchs []cpuArch
 	switch e.os {
 	case "linux":
 		for linuxArch, _ := range supportedLinuxArch {
@@ -703,6 +701,7 @@ func (bi *BinaryInfo) AddImage(path string, addr uint64) error {
 	if err != nil {
 		bi.Images[len(bi.Images)-1].loadErr = err
 	}
+	bi.macOSDebugFrameBugWorkaround()
 	return err
 }
 
@@ -808,7 +807,7 @@ func (bi *BinaryInfo) LoadImageFromData(dwdata *dwarf.Data, debugFrameBytes, deb
 	image.dwarfTreeCache, _ = simplelru.NewLRU(dwarfTreeCacheSize, nil)
 
 	if debugFrameBytes != nil {
-		bi.frameEntries = frame.Parse(debugFrameBytes, frame.DwarfEndian(debugFrameBytes), 0, bi.Arch.PtrSize())
+		bi.frameEntries, _ = frame.Parse(debugFrameBytes, frame.DwarfEndian(debugFrameBytes), 0, bi.Arch.PtrSize(), 0)
 	}
 
 	image.loclist2 = loclist.NewDwarf2Reader(debugLocBytes, bi.Arch.PtrSize())
@@ -1008,6 +1007,38 @@ func (bi *BinaryInfo) funcToImage(fn *Function) *Image {
 	return fn.cu.image
 }
 
+// parseDebugFrameGeneral parses a debug_frame and a eh_frame section.
+// At least one of the two must be present and parsed correctly, if
+// debug_frame is present it must be parsable correctly.
+func (bi *BinaryInfo) parseDebugFrameGeneral(image *Image, debugFrameBytes []byte, debugFrameName string, debugFrameErr error, ehFrameBytes []byte, ehFrameAddr uint64, ehFrameName string, byteOrder binary.ByteOrder) {
+	if debugFrameBytes == nil && ehFrameBytes == nil {
+		image.setLoadError("could not get %s section: %v", debugFrameName, debugFrameErr)
+		return
+	}
+
+	if debugFrameBytes != nil {
+		fe, err := frame.Parse(debugFrameBytes, byteOrder, image.StaticBase, bi.Arch.PtrSize(), 0)
+		if err != nil {
+			image.setLoadError("could not parse %s section: %v", debugFrameName, err)
+			return
+		}
+		bi.frameEntries = bi.frameEntries.Append(fe)
+	}
+
+	if ehFrameBytes != nil && ehFrameAddr > 0 {
+		fe, err := frame.Parse(ehFrameBytes, byteOrder, image.StaticBase, bi.Arch.PtrSize(), ehFrameAddr)
+		if err != nil {
+			if debugFrameBytes == nil {
+				image.setLoadError("could not parse %s section: %v", ehFrameName, err)
+				return
+			}
+			bi.logger.Warnf("could not parse %s section: %v", ehFrameName, err)
+			return
+		}
+		bi.frameEntries = bi.frameEntries.Append(fe)
+	}
+}
+
 // ELF ///////////////////////////////////////////////////////////////
 
 // ErrNoBuildIDNote is used in openSeparateDebugInfo to signal there's no
@@ -1184,8 +1215,8 @@ func loadBinaryInfoElf(bi *BinaryInfo, image *Image, path string, addr uint64, w
 	return nil
 }
 
-//  STT_FUNC is a code object, see /usr/include/elf.h for a full definition.
-const STT_FUNC = 2
+//  _STT_FUNC is a code object, see /usr/include/elf.h for a full definition.
+const _STT_FUNC = 2
 
 func (bi *BinaryInfo) loadSymbolName(image *Image, file *elf.File, wg *sync.WaitGroup) {
 	defer wg.Done()
@@ -1195,7 +1226,7 @@ func (bi *BinaryInfo) loadSymbolName(image *Image, file *elf.File, wg *sync.Wait
 	symSecs, _ := file.Symbols()
 	if symSecs != nil {
 		for _, symSec := range symSecs {
-			if symSec.Info == STT_FUNC { // TODO(chainhelen), need to parse others types.
+			if symSec.Info == _STT_FUNC { // TODO(chainhelen), need to parse others types.
 				s := symSec
 				bi.SymNames[symSec.Value+image.StaticBase] = &s
 			}
@@ -1206,13 +1237,16 @@ func (bi *BinaryInfo) loadSymbolName(image *Image, file *elf.File, wg *sync.Wait
 func (bi *BinaryInfo) parseDebugFrameElf(image *Image, exe *elf.File, debugInfoBytes []byte, wg *sync.WaitGroup) {
 	defer wg.Done()
 
-	debugFrameData, err := godwarf.GetDebugSectionElf(exe, "frame")
-	if err != nil {
-		image.setLoadError("could not get .debug_frame section: %v", err)
-		return
+	debugFrameData, debugFrameErr := godwarf.GetDebugSectionElf(exe, "frame")
+	ehFrameSection := exe.Section(".eh_frame")
+	var ehFrameData []byte
+	var ehFrameAddr uint64
+	if ehFrameSection != nil {
+		ehFrameAddr = ehFrameSection.Addr
+		ehFrameData, _ = ehFrameSection.Data()
 	}
 
-	bi.frameEntries = bi.frameEntries.Append(frame.Parse(debugFrameData, frame.DwarfEndian(debugInfoBytes), image.StaticBase, bi.Arch.PtrSize()))
+	bi.parseDebugFrameGeneral(image, debugFrameData, ".debug_frame", debugFrameErr, ehFrameData, ehFrameAddr, ".eh_frame", frame.DwarfEndian(debugInfoBytes))
 }
 
 func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.WaitGroup) {
@@ -1226,19 +1260,9 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 	// - Otherwise, Go asks the external linker to place the G pointer by
 	//   emitting runtime.tlsg, a TLS symbol, which is relocated to the chosen
 	//   offset in libc's TLS block.
-	symbols, err := exe.Symbols()
-	if err != nil {
-		image.setLoadError("could not parse ELF symbols: %v", err)
-		return
-	}
-	var tlsg *elf.Symbol
-	for _, symbol := range symbols {
-		if symbol.Name == "runtime.tlsg" {
-			s := symbol
-			tlsg = &s
-			break
-		}
-	}
+	// - On ARM64 (but really, any architecture other than i386 and 86x64) the
+	//   offset is calculate using runtime.tls_g and the formula is different.
+
 	var tls *elf.Prog
 	for _, prog := range exe.Progs {
 		if prog.Type == elf.PT_TLS {
@@ -1246,20 +1270,54 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 			break
 		}
 	}
-	if tlsg == nil || tls == nil {
-		bi.gStructOffset = ^uint64(bi.Arch.PtrSize()) + 1 //-ptrSize
-		return
+
+	switch exe.Machine {
+	case elf.EM_X86_64, elf.EM_386:
+		tlsg := getSymbol(image, exe, "runtime.tlsg")
+		if tlsg == nil || tls == nil {
+			bi.gStructOffset = ^uint64(bi.Arch.PtrSize()) + 1 //-ptrSize
+			return
+		}
+
+		// According to https://reviews.llvm.org/D61824, linkers must pad the actual
+		// size of the TLS segment to ensure that (tlsoffset%align) == (vaddr%align).
+		// This formula, copied from the lld code, matches that.
+		// https://github.com/llvm-mirror/lld/blob/9aef969544981d76bea8e4d1961d3a6980980ef9/ELF/InputSection.cpp#L643
+		memsz := tls.Memsz + (-tls.Vaddr-tls.Memsz)&(tls.Align-1)
+
+		// The TLS register points to the end of the TLS block, which is
+		// tls.Memsz long. runtime.tlsg is an offset from the beginning of that block.
+		bi.gStructOffset = ^(memsz) + 1 + tlsg.Value // -tls.Memsz + tlsg.Value
+
+	case elf.EM_AARCH64:
+		tlsg := getSymbol(image, exe, "runtime.tls_g")
+		if tlsg == nil || tls == nil {
+			bi.gStructOffset = 2 * uint64(bi.Arch.PtrSize())
+			return
+		}
+
+		bi.gStructOffset = tlsg.Value + uint64(bi.Arch.PtrSize()*2) + ((tls.Vaddr - uint64(bi.Arch.PtrSize()*2)) & (tls.Align - 1))
+
+	default:
+		// we should never get here
+		panic("architecture not supported")
+	}
+}
+
+func getSymbol(image *Image, exe *elf.File, name string) *elf.Symbol {
+	symbols, err := exe.Symbols()
+	if err != nil {
+		image.setLoadError("could not parse ELF symbols: %v", err)
+		return nil
 	}
 
-	// According to https://reviews.llvm.org/D61824, linkers must pad the actual
-	// size of the TLS segment to ensure that (tlsoffset%align) == (vaddr%align).
-	// This formula, copied from the lld code, matches that.
-	// https://github.com/llvm-mirror/lld/blob/9aef969544981d76bea8e4d1961d3a6980980ef9/ELF/InputSection.cpp#L643
-	memsz := tls.Memsz + (-tls.Vaddr-tls.Memsz)&(tls.Align-1)
-
-	// The TLS register points to the end of the TLS block, which is
-	// tls.Memsz long. runtime.tlsg is an offset from the beginning of that block.
-	bi.gStructOffset = ^(memsz) + 1 + tlsg.Value // -tls.Memsz + tlsg.Value
+	for _, symbol := range symbols {
+		if symbol.Name == name {
+			s := symbol
+			return &s
+		}
+	}
+	return nil
 }
 
 // PE ////////////////////////////////////////////////////////////////
@@ -1273,7 +1331,7 @@ func loadBinaryInfoPE(bi *BinaryInfo, image *Image, path string, entryPoint uint
 		return err
 	}
 	image.closer = closer
-	cpuArch := PEMachine(peFile.Machine)
+	cpuArch := _PEMachine(peFile.Machine)
 	if !supportedWindowsArch[cpuArch] {
 		return &ErrUnsupportedArch{os: "windows", cpuArch: cpuArch}
 	}
@@ -1338,29 +1396,7 @@ func (bi *BinaryInfo) parseDebugFramePE(image *Image, exe *pe.File, debugInfoByt
 	defer wg.Done()
 
 	debugFrameBytes, err := godwarf.GetDebugSectionPE(exe, "frame")
-	if err != nil {
-		image.setLoadError("could not get .debug_frame section: %v", err)
-		return
-	}
-
-	bi.frameEntries = bi.frameEntries.Append(frame.Parse(debugFrameBytes, frame.DwarfEndian(debugInfoBytes), image.StaticBase, bi.Arch.PtrSize()))
-}
-
-// Borrowed from https://golang.org/src/cmd/internal/objfile/pe.go
-func findPESymbol(f *pe.File, name string) (*pe.Symbol, error) {
-	for _, s := range f.Symbols {
-		if s.Name != name {
-			continue
-		}
-		if s.SectionNumber <= 0 {
-			return nil, fmt.Errorf("symbol %s: invalid section number %d", name, s.SectionNumber)
-		}
-		if len(f.Sections) < int(s.SectionNumber) {
-			return nil, fmt.Errorf("symbol %s: section number %d is larger than max %d", name, s.SectionNumber, len(f.Sections))
-		}
-		return s, nil
-	}
-	return nil, fmt.Errorf("no %s symbol found", name)
+	bi.parseDebugFrameGeneral(image, debugFrameBytes, ".debug_frame", err, nil, 0, "", frame.DwarfEndian(debugInfoBytes))
 }
 
 // MACH-O ////////////////////////////////////////////////////////////
@@ -1368,9 +1404,19 @@ func findPESymbol(f *pe.File, name string) (*pe.Symbol, error) {
 // loadBinaryInfoMacho specifically loads information from a Mach-O binary.
 func loadBinaryInfoMacho(bi *BinaryInfo, image *Image, path string, entryPoint uint64, wg *sync.WaitGroup) error {
 	exe, err := macho.Open(path)
+
 	if err != nil {
 		return err
 	}
+
+	if entryPoint != 0 {
+		// This is a little bit hacky. We use the entryPoint variable, but it
+		// actually holds the address of the mach-o header. We can use this
+		// to calculate the offset to the non-aslr location of the mach-o header
+		// (which is 0x100000000)
+		image.StaticBase = entryPoint - 0x100000000
+	}
+
 	image.closer = exe
 	if !supportedDarwinArch[exe.Cpu] {
 		return &ErrUnsupportedArch{os: "darwin", cpuArch: exe.Cpu}
@@ -1418,13 +1464,94 @@ func (bi *BinaryInfo) setGStructOffsetMacho() {
 func (bi *BinaryInfo) parseDebugFrameMacho(image *Image, exe *macho.File, debugInfoBytes []byte, wg *sync.WaitGroup) {
 	defer wg.Done()
 
-	debugFrameBytes, err := godwarf.GetDebugSectionMacho(exe, "frame")
-	if err != nil {
-		image.setLoadError("could not get __debug_frame section: %v", err)
+	debugFrameBytes, debugFrameErr := godwarf.GetDebugSectionMacho(exe, "frame")
+	ehFrameSection := exe.Section("__eh_frame")
+	var ehFrameBytes []byte
+	var ehFrameAddr uint64
+	if ehFrameSection != nil {
+		ehFrameAddr = ehFrameSection.Addr
+		ehFrameBytes, _ = ehFrameSection.Data()
+	}
+
+	bi.parseDebugFrameGeneral(image, debugFrameBytes, "__debug_frame", debugFrameErr, ehFrameBytes, ehFrameAddr, "__eh_frame", frame.DwarfEndian(debugInfoBytes))
+}
+
+// macOSDebugFrameBugWorkaround applies a workaround for:
+//  https://github.com/golang/go/issues/25841
+// It finds the Go function with the lowest entry point and the first
+// debug_frame FDE, calculates the difference between the start of the
+// function and the start of the FDE and sums it to all debug_frame FDEs.
+// A number of additional checks are performed to make sure we don't ruin
+// executables unaffected by this bug.
+func (bi *BinaryInfo) macOSDebugFrameBugWorkaround() {
+	//TODO: log extensively because of bugs in the field
+	if bi.GOOS != "darwin" || bi.Arch.Name != "arm64" {
+		return
+	}
+	if len(bi.Images) > 1 {
+		// Only do this for the first executable, but it might work for plugins as
+		// well if we had a way to distinguish where entries in bi.frameEntries
+		// come from
+		return
+	}
+	exe, ok := bi.Images[0].closer.(*macho.File)
+	if !ok {
+		return
+	}
+	if exe.Flags&macho.FlagPIE == 0 {
+		bi.logger.Infof("debug_frame workaround not needed: not a PIE (%#x)", exe.Flags)
 		return
 	}
 
-	bi.frameEntries = bi.frameEntries.Append(frame.Parse(debugFrameBytes, frame.DwarfEndian(debugInfoBytes), image.StaticBase, bi.Arch.PtrSize()))
+	// Find first Go function (first = lowest entry point)
+	var fn *Function
+	for i := range bi.Functions {
+		if bi.Functions[i].cu.isgo && bi.Functions[i].Entry > 0 {
+			fn = &bi.Functions[i]
+			break
+		}
+	}
+	if fn == nil {
+		bi.logger.Warn("debug_frame workaround not applied: could not find a Go function")
+		return
+	}
+
+	if fde, _ := bi.frameEntries.FDEForPC(fn.Entry); fde != nil {
+		// Function is covered, no need to apply workaround
+		bi.logger.Warnf("debug_frame workaround not applied: function %s (at %#x) covered by %#x-%#x", fn.Name, fn.Entry, fde.Begin(), fde.End())
+		return
+	}
+
+	// Find lowest FDE in debug_frame
+	var fde *frame.FrameDescriptionEntry
+	for i := range bi.frameEntries {
+		if bi.frameEntries[i].CIE.CIE_id == ^uint32(0) {
+			fde = bi.frameEntries[i]
+			break
+		}
+	}
+
+	if fde == nil {
+		bi.logger.Warnf("debug_frame workaround not applied because there are no debug_frame entries (%d)", len(bi.frameEntries))
+		return
+	}
+
+	fnsize := fn.End - fn.Entry
+
+	if fde.End()-fde.Begin() != fnsize || fde.Begin() > fn.Entry {
+		bi.logger.Warnf("debug_frame workaround not applied: function %s (at %#x-%#x) has a different size than the first FDE (%#x-%#x) (or the FDE starts after the function)", fn.Name, fn.Entry, fn.End, fde.Begin(), fde.End())
+		return
+	}
+
+	delta := fn.Entry - fde.Begin()
+
+	bi.logger.Infof("applying debug_frame workaround +%#x: function %s (at %#x-%#x) and FDE %#x-%#x", delta, fn.Name, fn.Entry, fn.End, fde.Begin(), fde.End())
+
+	for i := range bi.frameEntries {
+		if bi.frameEntries[i].CIE.CIE_id == ^uint32(0) {
+			bi.frameEntries[i].Translate(delta)
+		}
+	}
 }
 
 // Do not call this function directly it isn't able to deal correctly with package paths
@@ -1770,24 +1897,19 @@ func (bi *BinaryInfo) addAbstractSubprogram(entry *dwarf.Entry, ctxt *loadDebugI
 	name, ok := subprogramEntryName(entry, cu)
 	if !ok {
 		bi.logger.Warnf("reading debug_info: abstract subprogram without name at %#x", entry.Offset)
-		if entry.Children {
-			reader.SkipChildren()
-		}
-		return
-	}
-
-	fn := Function{
-		Name:   name,
-		offset: entry.Offset,
-		cu:     cu,
+		// In some cases clang produces abstract subprograms that do not have a
+		// name, but we should process them anyway.
 	}
 
 	if entry.Children {
 		bi.loadDebugInfoMapsInlinedCalls(ctxt, reader, cu)
 	}
 
-	bi.Functions = append(bi.Functions, fn)
-	ctxt.abstractOriginTable[entry.Offset] = len(bi.Functions) - 1
+	originIdx := ctxt.lookupAbstractOrigin(bi, entry.Offset)
+	fn := &bi.Functions[originIdx]
+	fn.Name = name
+	fn.offset = entry.Offset
+	fn.cu = cu
 }
 
 // addConcreteInlinedSubprogram adds the concrete entry of a subprogram that was also inlined.
@@ -1801,19 +1923,12 @@ func (bi *BinaryInfo) addConcreteInlinedSubprogram(entry *dwarf.Entry, originOff
 		return
 	}
 
-	originIdx, ok := ctxt.abstractOriginTable[originOffset]
-	if !ok {
-		bi.logger.Warnf("reading debug_info: could not find abstract origin of concrete inlined subprogram at %#x (origin offset %#x)", entry.Offset, originOffset)
-		if entry.Children {
-			reader.SkipChildren()
-		}
-		return
-	}
-
+	originIdx := ctxt.lookupAbstractOrigin(bi, originOffset)
 	fn := &bi.Functions[originIdx]
 	fn.offset = entry.Offset
 	fn.Entry = lowpc
 	fn.End = highpc
+	fn.cu = cu
 
 	if entry.Children {
 		bi.loadDebugInfoMapsInlinedCalls(ctxt, reader, cu)
@@ -1826,29 +1941,25 @@ func (bi *BinaryInfo) addConcreteSubprogram(entry *dwarf.Entry, ctxt *loadDebugI
 	lowpc, highpc, ok := subprogramEntryRange(entry, cu.image)
 	if !ok {
 		bi.logger.Warnf("reading debug_info: concrete subprogram without address range at %#x", entry.Offset)
-		if entry.Children {
-			reader.SkipChildren()
-		}
-		return
+		// When clang inlines a function, in some cases, it produces a concrete
+		// subprogram without address range and then inlined calls that reference
+		// it, instead of producing an abstract subprogram.
+		// It is unclear if this behavior is standard.
 	}
 
 	name, ok := subprogramEntryName(entry, cu)
 	if !ok {
 		bi.logger.Warnf("reading debug_info: concrete subprogram without name at %#x", entry.Offset)
-		if entry.Children {
-			reader.SkipChildren()
-		}
-		return
 	}
 
-	fn := Function{
-		Name:   name,
-		Entry:  lowpc,
-		End:    highpc,
-		offset: entry.Offset,
-		cu:     cu,
-	}
-	bi.Functions = append(bi.Functions, fn)
+	originIdx := ctxt.lookupAbstractOrigin(bi, entry.Offset)
+	fn := &bi.Functions[originIdx]
+
+	fn.Name = name
+	fn.Entry = lowpc
+	fn.End = highpc
+	fn.offset = entry.Offset
+	fn.cu = cu
 
 	if entry.Children {
 		bi.loadDebugInfoMapsInlinedCalls(ctxt, reader, cu)
@@ -1894,14 +2005,6 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 				continue
 			}
 
-			originIdx, ok := ctxt.abstractOriginTable[originOffset]
-			if !ok {
-				bi.logger.Warnf("reading debug_info: could not find abstract origin (%#x) of inlined call at %#x", originOffset, entry.Offset)
-				reader.SkipChildren()
-				continue
-			}
-			fn := &bi.Functions[originIdx]
-
 			lowpc, highpc, ok := subprogramEntryRange(entry, cu.image)
 			if !ok {
 				bi.logger.Warnf("reading debug_info: inlined call without address range at %#x", entry.Offset)
@@ -1916,17 +2019,15 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 				reader.SkipChildren()
 				continue
 			}
-			if cu.lineInfo == nil {
-				bi.logger.Warnf("reading debug_info: inlined call on a compilation unit without debug_line section at %#x", entry.Offset)
+			callfile, cferr := cu.filePath(int(callfileidx), entry)
+			if cferr != nil {
+				bi.logger.Warnf("%v", cferr)
 				reader.SkipChildren()
 				continue
 			}
-			if int(callfileidx-1) >= len(cu.lineInfo.FileNames) {
-				bi.logger.Warnf("reading debug_info: CallFile (%d) of inlined call does not exist in compile unit file table at %#x", callfileidx, entry.Offset)
-				reader.SkipChildren()
-				continue
-			}
-			callfile := cu.lineInfo.FileNames[callfileidx-1].Path
+
+			originIdx := ctxt.lookupAbstractOrigin(bi, originOffset)
+			fn := &bi.Functions[originIdx]
 
 			fn.InlinedCalls = append(fn.InlinedCalls, InlinedCall{
 				cu:     cu,
@@ -1934,6 +2035,10 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 				HighPC: highpc,
 			})
 
+			if fn.cu == nil {
+				fn.cu = cu
+			}
+
 			fl := fileLine{callfile, int(callline)}
 			bi.inlinedCallLines[fl] = append(bi.inlinedCallLines[fl], lowpc)
 		}
@@ -2087,3 +2192,25 @@ func (bi *BinaryInfo) ListPackagesBuildInfo(includeFiles bool) []*PackageBuildIn
 	sort.Slice(r, func(i, j int) bool { return r[i].ImportPath < r[j].ImportPath })
 	return r
 }
+
+// cuFilePath takes a compilation unit "cu" and a file index reference
+// "fileidx" and returns the corresponding file name entry from the
+// DWARF line table associated with the unit; "entry" is the offset of
+// the attribute where the file reference originated, for logging
+// purposes. Return value is the file string and an error value; error
+// will be non-nil if the file could not be recovered, perhaps due to
+// malformed DWARF.
+func (cu *compileUnit) filePath(fileidx int, entry *dwarf.Entry) (string, error) {
+	if cu.lineInfo == nil {
+		return "", fmt.Errorf("reading debug_info: file reference within a compilation unit without debug_line section at %#x", entry.Offset)
+	}
+	// File numbering is slightly different before and after DWARF 5;
+	// account for this here. See section 6.2.4 of the DWARF 5 spec.
+	if cu.Version < 5 {
+		fileidx--
+	}
+	if fileidx < 0 || fileidx >= len(cu.lineInfo.FileNames) {
+		return "", fmt.Errorf("reading debug_info: file index (%d) out of range in compile unit file table at %#x", fileidx, entry.Offset)
+	}
+	return cu.lineInfo.FileNames[fileidx].Path, nil
+}
