@@ -3,7 +3,6 @@ package proc
 import (
 	"bytes"
 	"debug/dwarf"
-	"encoding/binary"
 	"errors"
 	"fmt"
 	"go/ast"
@@ -36,8 +35,6 @@ type EvalScope struct {
 
 	frameOffset int64
 
-	aordr *dwarf.Reader // extra reader to load DW_AT_abstract_origin entries, do not initialize
-
 	// When the following pointer is not nil this EvalScope was created
 	// by CallFunction and the expression evaluation is executing on a
 	// different goroutine from the debugger's main goroutine.
@@ -558,14 +555,6 @@ func (scope *EvalScope) image() *Image {
 	return scope.BinInfo.funcToImage(scope.Fn)
 }
 
-// globalFor returns a global scope for 'image' with the register values of 'scope'.
-func (scope *EvalScope) globalFor(image *Image) *EvalScope {
-	r := *scope
-	r.Regs.StaticBase = image.StaticBase
-	r.Fn = &Function{cu: &compileUnit{image: image}}
-	return &r
-}
-
 // DwarfReader returns the DwarfReader containing the
 // Dwarf information for the target process.
 func (scope *EvalScope) DwarfReader() *reader.Reader {
@@ -915,17 +904,14 @@ func (scope *EvalScope) evalTypeCast(node *ast.CallExpr) (*Variable, error) {
 }
 
 func convertInt(n uint64, signed bool, size int64) uint64 {
-	buf := make([]byte, 64/8)
-	binary.BigEndian.PutUint64(buf, n)
-	m := 64/8 - int(size)
-	s := byte(0)
-	if signed && (buf[m]&0x80 > 0) {
-		s = 0xff
+	bits := uint64(size) * 8
+	mask := uint64((1 << bits) - 1)
+	r := n & mask
+	if signed && (r>>(bits-1)) != 0 {
+		// sign extension
+		r |= ^uint64(0) &^ mask
 	}
-	for i := 0; i < m; i++ {
-		buf[i] = s
-	}
-	return uint64(binary.BigEndian.Uint64(buf))
+	return r
 }
 
 func (scope *EvalScope) evalBuiltinCall(node *ast.CallExpr) (*Variable, error) {
@@ -1152,6 +1138,36 @@ func (scope *EvalScope) evalIdent(node *ast.Ident) (*Variable, error) {
 			return v, nil
 		}
 	}
+
+	// not a local variable, nor a global variable, try a CPU register
+	if s := validRegisterName(node.Name); s != "" {
+		if regnum, ok := scope.BinInfo.Arch.RegisterNameToDwarf(s); ok {
+			if reg := scope.Regs.Reg(uint64(regnum)); reg != nil {
+				reg.FillBytes()
+
+				var typ godwarf.Type
+				if len(reg.Bytes) <= 8 {
+					typ = &godwarf.UintType{BasicType: godwarf.BasicType{CommonType: godwarf.CommonType{ByteSize: 8, Name: "uint64"}, BitSize: 64, BitOffset: 0}}
+				} else {
+					typ, err = scope.BinInfo.findType("string")
+					if err != nil {
+						return nil, err
+					}
+				}
+
+				v := newVariable(node.Name, 0, typ, scope.BinInfo, scope.Mem)
+				if v.Kind == reflect.String {
+					v.Len = int64(len(reg.Bytes) * 2)
+					v.Base = fakeAddress
+				}
+				v.Addr = fakeAddress
+				v.Flags = VariableCPURegister
+				v.reg = reg
+				return v, nil
+			}
+		}
+	}
+
 	return nil, fmt.Errorf("could not find symbol value for %s", node.Name)
 }
 
@@ -1170,6 +1186,10 @@ func (scope *EvalScope) evalStructSelector(node *ast.SelectorExpr) (*Variable, e
 	if xv.Addr == 0 && xv.Name == "" && xv.DwarfType == nil && xv.RealType == nil {
 		return nil, fmt.Errorf("%s (type %s) is not a struct", xv.Value, xv.TypeString())
 	}
+	// Special type conversions for CPU register variables (REGNAME.int8, etc)
+	if xv.Flags&VariableCPURegister != 0 && !xv.loaded {
+		return xv.registerVariableTypeConv(node.Sel.Name)
+	}
 
 	rv, err := xv.findMethod(node.Sel.Name)
 	if err != nil {
@@ -1612,8 +1632,15 @@ func (scope *EvalScope) evalBinary(node *ast.BinaryExpr) (*Variable, error) {
 
 		r := xv.newVariable("", 0, typ, scope.Mem)
 		r.Value = rc
-		if r.Kind == reflect.String {
+		switch r.Kind {
+		case reflect.String:
 			r.Len = xv.Len + yv.Len
+		case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
+			n, _ := constant.Int64Val(r.Value)
+			r.Value = constant.MakeInt64(int64(convertInt(uint64(n), true, typ.Size())))
+		case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
+			n, _ := constant.Uint64Val(r.Value)
+			r.Value = constant.MakeUint64(convertInt(n, false, typ.Size()))
 		}
 		return r, nil
 	}
@@ -1879,6 +1906,9 @@ func (v *Variable) sliceAccess(idx int) (*Variable, error) {
 	if wrong {
 		return nil, fmt.Errorf("index out of bounds")
 	}
+	if v.loaded {
+		return &v.Children[idx], nil
+	}
 	mem := v.mem
 	if v.Kind != reflect.Array {
 		mem = DereferenceMemory(mem)
@@ -1959,6 +1989,8 @@ func (v *Variable) reslice(low int64, high int64) (*Variable, error) {
 	r.Base = base
 	r.stride = v.stride
 	r.fieldType = v.fieldType
+	r.Flags = v.Flags
+	r.reg = v.reg
 
 	return r, nil
 }
@@ -1973,73 +2005,79 @@ func (v *Variable) findMethod(mname string) (*Variable, error) {
 		return v.Children[0].findMethod(mname)
 	}
 
-	typ := v.DwarfType
-	ptyp, isptr := typ.(*godwarf.PtrType)
-	if isptr {
-		typ = ptyp.Type
-	}
+	queue := []*Variable{v}
+	seen := map[string]struct{}{}
 
-	typePath := typ.Common().Name
-	dot := strings.LastIndex(typePath, ".")
-	if dot < 0 {
-		// probably just a C type
-		return nil, nil
-	}
-
-	pkg := typePath[:dot]
-	receiver := typePath[dot+1:]
-
-	if fn, ok := v.bi.LookupFunc[fmt.Sprintf("%s.%s.%s", pkg, receiver, mname)]; ok {
-		r, err := functionToVariable(fn, v.bi, v.mem)
-		if err != nil {
-			return nil, err
+	for len(queue) > 0 {
+		v := queue[0]
+		queue = append(queue[:0], queue[1:]...)
+		if _, isseen := seen[v.RealType.String()]; isseen {
+			continue
 		}
+		seen[v.RealType.String()] = struct{}{}
+
+		typ := v.DwarfType
+		ptyp, isptr := typ.(*godwarf.PtrType)
 		if isptr {
-			r.Children = append(r.Children, *(v.maybeDereference()))
-		} else {
-			r.Children = append(r.Children, *v)
+			typ = ptyp.Type
 		}
-		return r, nil
-	}
 
-	if fn, ok := v.bi.LookupFunc[fmt.Sprintf("%s.(*%s).%s", pkg, receiver, mname)]; ok {
-		r, err := functionToVariable(fn, v.bi, v.mem)
-		if err != nil {
-			return nil, err
+		typePath := typ.Common().Name
+		dot := strings.LastIndex(typePath, ".")
+		if dot < 0 {
+			// probably just a C type
+			continue
 		}
-		if isptr {
-			r.Children = append(r.Children, *v)
-		} else {
-			r.Children = append(r.Children, *(v.pointerToVariable()))
-		}
-		return r, nil
-	}
-	return v.tryFindMethodInEmbeddedFields(mname)
-}
 
-func (v *Variable) tryFindMethodInEmbeddedFields(mname string) (*Variable, error) {
-	structVar := v.maybeDereference()
-	structVar.Name = v.Name
-	if structVar.Unreadable != nil {
-		return structVar, nil
-	}
-	switch t := structVar.RealType.(type) {
-	case *godwarf.StructType:
-		for _, field := range t.Field {
-			if field.Embedded {
-				// Recursively check for promoted fields on the embedded field
-				embeddedVar, err := structVar.toField(field)
-				if err != nil {
-					return nil, err
-				}
-				if embeddedMethod, err := embeddedVar.findMethod(mname); err != nil {
-					return nil, err
-				} else if embeddedMethod != nil {
-					return embeddedMethod, nil
+		pkg := typePath[:dot]
+		receiver := typePath[dot+1:]
+
+		if fn, ok := v.bi.LookupFunc[fmt.Sprintf("%s.%s.%s", pkg, receiver, mname)]; ok {
+			r, err := functionToVariable(fn, v.bi, v.mem)
+			if err != nil {
+				return nil, err
+			}
+			if isptr {
+				r.Children = append(r.Children, *(v.maybeDereference()))
+			} else {
+				r.Children = append(r.Children, *v)
+			}
+			return r, nil
+		}
+
+		if fn, ok := v.bi.LookupFunc[fmt.Sprintf("%s.(*%s).%s", pkg, receiver, mname)]; ok {
+			r, err := functionToVariable(fn, v.bi, v.mem)
+			if err != nil {
+				return nil, err
+			}
+			if isptr {
+				r.Children = append(r.Children, *v)
+			} else {
+				r.Children = append(r.Children, *(v.pointerToVariable()))
+			}
+			return r, nil
+		}
+
+		// queue embedded fields for search
+		structVar := v.maybeDereference()
+		structVar.Name = v.Name
+		if structVar.Unreadable != nil {
+			return structVar, nil
+		}
+		switch t := structVar.RealType.(type) {
+		case *godwarf.StructType:
+			for _, field := range t.Field {
+				if field.Embedded {
+					embeddedVar, err := structVar.toField(field)
+					if err != nil {
+						return nil, err
+					}
+					queue = append(queue, embeddedVar)
 				}
 			}
 		}
 	}
+
 	return nil, nil
 }
 
@@ -2158,3 +2196,15 @@ func (fn *Function) fakeType(bi *BinaryInfo, removeReceiver bool) (*godwarf.Func
 		// If we start using them they should be filled here.
 	}, nil
 }
+
+func validRegisterName(s string) string {
+	for len(s) > 0 && s[0] == '_' {
+		s = s[1:]
+	}
+	for i := range s {
+		if (s[i] < '0' || s[i] > '9') && (s[i] < 'A' || s[i] > 'Z') {
+			return ""
+		}
+	}
+	return s
+}
