@@ -74,10 +74,11 @@ import (
 	"runtime"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 
-	"golang.org/x/arch/x86/x86asm"
-
+	"github.com/go-delve/delve/pkg/dwarf/op"
+	"github.com/go-delve/delve/pkg/elfwriter"
 	"github.com/go-delve/delve/pkg/logflags"
 	"github.com/go-delve/delve/pkg/proc"
 	"github.com/go-delve/delve/pkg/proc/linutil"
@@ -90,6 +91,8 @@ const (
 
 	maxTransmitAttempts    = 3    // number of retransmission attempts on failed checksum
 	initialInputBufferSize = 2048 // size of the input buffer for gdbConn
+
+	debugServerEnvVar = "DELVE_DEBUGSERVER_PATH" // use this environment variable to override the path to debugserver used by Launch/Attach
 )
 
 const heartbeatInterval = 10 * time.Second
@@ -126,11 +129,15 @@ var debugserverExecutablePaths = []string{
 // while there are still internal breakpoints set.
 var ErrDirChange = errors.New("direction change with internal breakpoints")
 
+var checkCanUnmaskSignalsOnce sync.Once
+var canUnmaskSignalsCached bool
+
 // gdbProcess implements proc.Process using a connection to a debugger stub
 // that understands Gdb Remote Serial Protocol.
 type gdbProcess struct {
-	bi   *proc.BinaryInfo
-	conn gdbConn
+	bi       *proc.BinaryInfo
+	regnames *gdbRegnames
+	conn     gdbConn
 
 	threads       map[int]*gdbThread
 	currentThread *gdbThread
@@ -148,6 +155,8 @@ type gdbProcess struct {
 
 	loadGInstrAddr uint64 // address of the g loading instruction, zero if we couldn't allocate it
 
+	breakpointKind int // breakpoint kind to pass to 'z' and 'Z' when creating software breakpoints
+
 	process  *os.Process
 	waitChan chan *os.ProcessState
 
@@ -186,6 +195,8 @@ type gdbRegisters struct {
 	gaddr    uint64
 	hasgaddr bool
 	buf      []byte
+	arch     *proc.Arch
+	regnames *gdbRegnames
 }
 
 type gdbRegister struct {
@@ -193,6 +204,11 @@ type gdbRegister struct {
 	regnum int
 }
 
+// gdbRegname records names of important CPU registers
+type gdbRegnames struct {
+	PC, SP, BP, CX, FsBase string
+}
+
 // newProcess creates a new Process instance.
 // If process is not nil it is the stub's process and will be killed after
 // Detach.
@@ -208,12 +224,37 @@ func newProcess(process *os.Process) *gdbProcess {
 		},
 		threads:        make(map[int]*gdbThread),
 		bi:             proc.NewBinaryInfo(runtime.GOOS, runtime.GOARCH),
+		regnames:       new(gdbRegnames),
 		breakpoints:    proc.NewBreakpointMap(),
 		gcmdok:         true,
 		threadStopInfo: true,
 		process:        process,
 	}
 
+	switch p.bi.Arch.Name {
+	default:
+		fallthrough
+	case "amd64":
+		p.breakpointKind = 1
+	case "arm64":
+		p.breakpointKind = 4
+	}
+
+	p.regnames.PC = registerName(p.bi.Arch, p.bi.Arch.PCRegNum)
+	p.regnames.SP = registerName(p.bi.Arch, p.bi.Arch.SPRegNum)
+	p.regnames.BP = registerName(p.bi.Arch, p.bi.Arch.BPRegNum)
+
+	switch p.bi.Arch.Name {
+	case "arm64":
+		p.regnames.BP = "fp"
+		p.regnames.CX = "x0"
+	case "amd64":
+		p.regnames.CX = "rcx"
+		p.regnames.FsBase = "fs_base"
+	default:
+		panic("not implemented")
+	}
+
 	if process != nil {
 		p.waitChan = make(chan *os.ProcessState)
 		go func() {
@@ -272,7 +313,7 @@ func (p *gdbProcess) Dial(addr string, path string, pid int, debugInfoDirs []str
 func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs []string, stopReason proc.StopReason) (*proc.Target, error) {
 	p.conn.conn = conn
 	p.conn.pid = pid
-	err := p.conn.handshake()
+	err := p.conn.handshake(p.regnames)
 	if err != nil {
 		conn.Close()
 		return nil, err
@@ -286,6 +327,16 @@ func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs
 					p.gcmdok = false
 					break
 				}
+
+				// Workaround for darwin arm64. Apple's debugserver seems to have a problem
+				// with not selecting the correct thread in the 'g' command and the returned
+				// registers are empty / an E74 error is thrown. This was reported to LLVM
+				// as https://bugs.llvm.org/show_bug.cgi?id=50169
+				version, err := strconv.ParseInt(v[len("version:"):], 10, 32)
+
+				if err == nil && version >= 1200 && version <= 1205 && p.bi.Arch.Name == "arm64" {
+					p.gcmdok = false
+				}
 			}
 		}
 	}
@@ -295,17 +346,19 @@ func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs
 		return nil, err
 	}
 
-	// None of the stubs we support returns the value of fs_base or gs_base
-	// along with the registers, therefore we have to resort to executing a MOV
-	// instruction on the inferior to find out where the G struct of a given
-	// thread is located.
-	// Here we try to allocate some memory on the inferior which we will use to
-	// store the MOV instruction.
-	// If the stub doesn't support memory allocation reloadRegisters will
-	// overwrite some existing memory to store the MOV.
-	if addr, err := p.conn.allocMemory(256); err == nil {
-		if _, err := p.conn.writeMemory(addr, p.loadGInstr()); err == nil {
-			p.loadGInstrAddr = addr
+	if p.bi.Arch.Name != "arm64" {
+		// None of the stubs we support returns the value of fs_base or gs_base
+		// along with the registers, therefore we have to resort to executing a MOV
+		// instruction on the inferior to find out where the G struct of a given
+		// thread is located.
+		// Here we try to allocate some memory on the inferior which we will use to
+		// store the MOV instruction.
+		// If the stub doesn't support memory allocation reloadRegisters will
+		// overwrite some existing memory to store the MOV.
+		if addr, err := p.conn.allocMemory(256); err == nil {
+			if _, err := p.conn.writeMemory(addr, p.loadGInstr()); err == nil {
+				p.loadGInstrAddr = addr
+			}
 		}
 	}
 
@@ -330,6 +383,9 @@ func unusedPort() string {
 // getDebugServerAbsolutePath returns a string of the absolute path to the debugserver binary IFF it is
 // found in the system path ($PATH), the Xcode bundle or the standalone CLT location.
 func getDebugServerAbsolutePath() string {
+	if path := os.Getenv(debugServerEnvVar); path != "" {
+		return path
+	}
 	for _, debugServerPath := range debugserverExecutablePaths {
 		if debugServerPath == "" {
 			continue
@@ -341,6 +397,14 @@ func getDebugServerAbsolutePath() string {
 	return ""
 }
 
+func canUnmaskSignals(debugServerExecutable string) bool {
+	checkCanUnmaskSignalsOnce.Do(func() {
+		buf, _ := exec.Command(debugServerExecutable, "--unmask-singals").CombinedOutput()
+		canUnmaskSignalsCached = !strings.Contains(string(buf), "unrecognized option")
+	})
+	return canUnmaskSignalsCached
+}
+
 // commandLogger is a wrapper around the exec.Command() function to log the arguments prior to
 // starting the process
 func commandLogger(binary string, arguments ...string) *exec.Cmd {
@@ -421,6 +485,9 @@ func LLDBLaunch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs [
 		if flags&proc.LaunchDisableASLR != 0 {
 			args = append(args, "-D")
 		}
+		if canUnmaskSignals(debugserverExecutable) {
+			args = append(args, "--unmask-signals")
+		}
 		args = append(args, "-F", "-R", fmt.Sprintf("127.0.0.1:%d", listener.Addr().(*net.TCPAddr).Port), "--")
 		args = append(args, cmd...)
 
@@ -500,7 +567,11 @@ func LLDBAttach(pid int, path string, debugInfoDirs []string) (*proc.Target, err
 		if err != nil {
 			return nil, err
 		}
-		process = commandLogger(debugserverExecutable, "-R", fmt.Sprintf("127.0.0.1:%d", listener.Addr().(*net.TCPAddr).Port), "--attach="+strconv.Itoa(pid))
+		args := []string{"-R", fmt.Sprintf("127.0.0.1:%d", listener.Addr().(*net.TCPAddr).Port), "--attach=" + strconv.Itoa(pid)}
+		if canUnmaskSignals(debugserverExecutable) {
+			args = append(args, "--unmask-signals")
+		}
+		process = commandLogger(debugserverExecutable, args...)
 	} else {
 		if _, err = exec.LookPath("lldb-server"); err != nil {
 			return nil, &ErrBackendUnavailable{}
@@ -533,12 +604,27 @@ func LLDBAttach(pid int, path string, debugInfoDirs []string) (*proc.Target, err
 // debugging PIEs.
 func (p *gdbProcess) EntryPoint() (uint64, error) {
 	var entryPoint uint64
-	if auxv, err := p.conn.readAuxv(); err == nil {
+	if p.bi.GOOS == "darwin" && p.bi.Arch.Name == "arm64" {
+		// There is no auxv on darwin, however, we can get the location of the mach-o
+		// header from the debugserver by going through the loaded libraries, which includes
+		// the exe itself
+		images, _ := p.conn.getLoadedDynamicLibraries()
+		for _, image := range images {
+			if image.MachHeader.FileType == macho.TypeExec {
+				// This is a bit hacky. This is technically not the entrypoint,
+				// but rather we use the variable to points at the mach-o header,
+				// so we can get the offset in bininfo
+				entryPoint = image.LoadAddress
+				break
+			}
+		}
+	} else if auxv, err := p.conn.readAuxv(); err == nil {
 		// If we can't read the auxiliary vector it just means it's not supported
 		// by the OS or by the stub. If we are debugging a PIE and the entry point
 		// is needed proc.LoadBinaryInfo will complain about it.
 		entryPoint = linutil.EntryPointFromAuxv(auxv, p.BinInfo().Arch.PtrSize())
 	}
+
 	return entryPoint, nil
 }
 
@@ -600,7 +686,8 @@ func (p *gdbProcess) initialize(path string, debugInfoDirs []string, stopReason
 		Path:                path,
 		DebugInfoDirs:       debugInfoDirs,
 		DisableAsyncPreempt: runtime.GOOS == "darwin",
-		StopReason:          stopReason})
+		StopReason:          stopReason,
+		CanDump:             runtime.GOOS == "darwin"})
 	if err != nil {
 		p.conn.conn.Close()
 		return nil, err
@@ -643,7 +730,7 @@ func (p *gdbProcess) Valid() (bool, error) {
 		return false, proc.ErrProcessDetached
 	}
 	if p.exited {
-		return false, &proc.ErrProcessExited{Pid: p.Pid()}
+		return false, proc.ErrProcessExited{Pid: p.Pid()}
 	}
 	return true, nil
 }
@@ -693,7 +780,7 @@ const (
 // a breakpoint is hit or signal is received.
 func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
 	if p.exited {
-		return nil, proc.StopExited, &proc.ErrProcessExited{Pid: p.conn.pid}
+		return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
 	}
 
 	if p.conn.direction == proc.Forward {
@@ -985,7 +1072,7 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 	p.clearThreadRegisters()
 
 	for addr := range p.breakpoints.M {
-		p.conn.setBreakpoint(addr)
+		p.conn.setBreakpoint(addr, p.breakpointKind)
 	}
 
 	return p.currentThread, p.setCurrentBreakpoints()
@@ -1121,18 +1208,15 @@ func (p *gdbProcess) FindBreakpoint(pc uint64) (*proc.Breakpoint, bool) {
 	return nil, false
 }
 
-func (p *gdbProcess) WriteBreakpoint(addr uint64) (string, int, *proc.Function, []byte, error) {
-	f, l, fn := p.bi.PCToLine(uint64(addr))
-
-	if err := p.conn.setBreakpoint(addr); err != nil {
-		return "", 0, nil, nil, err
+func (p *gdbProcess) WriteBreakpoint(bp *proc.Breakpoint) error {
+	if bp.WatchType != 0 {
+		return errors.New("hardware breakpoints not supported")
 	}
-
-	return f, l, fn, nil, nil
+	return p.conn.setBreakpoint(bp.Addr, p.breakpointKind)
 }
 
 func (p *gdbProcess) EraseBreakpoint(bp *proc.Breakpoint) error {
-	return p.conn.clearBreakpoint(bp.Addr)
+	return p.conn.clearBreakpoint(bp.Addr, p.breakpointKind)
 }
 
 type threadUpdater struct {
@@ -1303,7 +1387,7 @@ func (t *gdbThread) Location() (*proc.Location, error) {
 	if err != nil {
 		return nil, err
 	}
-	if pcreg, ok := regs.(*gdbRegisters).regs[regnamePC]; !ok {
+	if pcreg, ok := regs.(*gdbRegisters).regs[regs.(*gdbRegisters).regnames.PC]; !ok {
 		t.p.conn.log.Errorf("thread %d could not find RIP register", t.ID)
 	} else if len(pcreg.value) < t.p.bi.Arch.PtrSize() {
 		t.p.conn.log.Errorf("thread %d bad length for RIP register: %d", t.ID, len(pcreg.value))
@@ -1353,11 +1437,11 @@ func (t *gdbThread) Common() *proc.CommonThread {
 func (t *gdbThread) StepInstruction() error {
 	pc := t.regs.PC()
 	if _, atbp := t.p.breakpoints.M[pc]; atbp {
-		err := t.p.conn.clearBreakpoint(pc)
+		err := t.p.conn.clearBreakpoint(pc, t.p.breakpointKind)
 		if err != nil {
 			return err
 		}
-		defer t.p.conn.setBreakpoint(pc)
+		defer t.p.conn.setBreakpoint(pc, t.p.breakpointKind)
 	}
 	// Reset thread registers so the next call to
 	// Thread.Registers will not be cached.
@@ -1412,7 +1496,40 @@ func (p *gdbProcess) loadGInstr() []byte {
 	return buf.Bytes()
 }
 
-func (regs *gdbRegisters) init(regsInfo []gdbRegisterInfo) {
+func (p *gdbProcess) MemoryMap() ([]proc.MemoryMapEntry, error) {
+	r := []proc.MemoryMapEntry{}
+	addr := uint64(0)
+	for addr != ^uint64(0) {
+		mri, err := p.conn.memoryRegionInfo(addr)
+		if err != nil {
+			return nil, err
+		}
+		if addr+mri.size <= addr {
+			return nil, errors.New("qMemoryRegionInfo response wrapped around the address space or stuck")
+		}
+		if mri.permissions != "" {
+			var mme proc.MemoryMapEntry
+
+			mme.Addr = addr
+			mme.Size = mri.size
+			mme.Read = strings.Contains(mri.permissions, "r")
+			mme.Write = strings.Contains(mri.permissions, "w")
+			mme.Exec = strings.Contains(mri.permissions, "x")
+
+			r = append(r, mme)
+		}
+		addr += mri.size
+	}
+	return r, nil
+}
+
+func (p *gdbProcess) DumpProcessNotes(notes []elfwriter.Note, threadDone func()) (threadsDone bool, out []elfwriter.Note, err error) {
+	return false, notes, nil
+}
+
+func (regs *gdbRegisters) init(regsInfo []gdbRegisterInfo, arch *proc.Arch, regnames *gdbRegnames) {
+	regs.arch = arch
+	regs.regnames = regnames
 	regs.regs = make(map[string]gdbRegister)
 	regs.regsInfo = regsInfo
 
@@ -1434,12 +1551,13 @@ func (regs *gdbRegisters) init(regsInfo []gdbRegisterInfo) {
 // the stub can allocate memory, or reloadGAtPC, if the stub can't.
 func (t *gdbThread) reloadRegisters() error {
 	if t.regs.regs == nil {
-		t.regs.init(t.p.conn.regsInfo)
+		t.regs.init(t.p.conn.regsInfo, t.p.bi.Arch, t.p.regnames)
 	}
 
 	if t.p.gcmdok {
 		if err := t.p.conn.readRegisters(t.strID, t.regs.buf); err != nil {
-			if isProtocolErrorUnsupported(err) {
+			gdberr, isProt := err.(*GdbProtocolError)
+			if isProtocolErrorUnsupported(err) || (t.p.conn.isDebugserver && isProt && gdberr.code == "E74") {
 				t.p.gcmdok = false
 			} else {
 				return err
@@ -1456,7 +1574,7 @@ func (t *gdbThread) reloadRegisters() error {
 
 	switch t.p.bi.GOOS {
 	case "linux":
-		if reg, hasFsBase := t.regs.regs[regnameFsBase]; hasFsBase {
+		if reg, hasFsBase := t.regs.regs[t.p.regnames.FsBase]; hasFsBase {
 			t.regs.gaddr = 0
 			t.regs.tls = binary.LittleEndian.Uint64(reg.value)
 			t.regs.hasgaddr = false
@@ -1464,10 +1582,21 @@ func (t *gdbThread) reloadRegisters() error {
 		}
 	}
 
-	if t.p.loadGInstrAddr > 0 {
-		return t.reloadGAlloc()
+	if t.p.bi.Arch.Name == "arm64" {
+		// no need to play around with the GInstr on ARM64 because
+		// the G addr is stored in a register
+
+		t.regs.gaddr = t.regs.byName("x28")
+		t.regs.hasgaddr = true
+		t.regs.tls = 0
+	} else {
+		if t.p.loadGInstrAddr > 0 {
+			return t.reloadGAlloc()
+		}
+		return t.reloadGAtPC()
 	}
-	return t.reloadGAtPC()
+
+	return nil
 }
 
 func (t *gdbThread) writeSomeRegisters(regNames ...string) error {
@@ -1532,11 +1661,11 @@ func (t *gdbThread) reloadGAtPC() error {
 	// Additionally all breakpoints in [pc, pc+len(movinstr)] need to be removed
 	for addr := range t.p.breakpoints.M {
 		if addr >= pc && addr <= pc+uint64(len(movinstr)) {
-			err := t.p.conn.clearBreakpoint(addr)
+			err := t.p.conn.clearBreakpoint(addr, t.p.breakpointKind)
 			if err != nil {
 				return err
 			}
-			defer t.p.conn.setBreakpoint(addr)
+			defer t.p.conn.setBreakpoint(addr, t.p.breakpointKind)
 		}
 	}
 
@@ -1558,7 +1687,7 @@ func (t *gdbThread) reloadGAtPC() error {
 		}
 		t.regs.setPC(pc)
 		t.regs.setCX(cx)
-		err1 := t.writeSomeRegisters(regnamePC, regnameCX)
+		err1 := t.writeSomeRegisters(t.p.regnames.PC, t.p.regnames.CX)
 		if err == nil {
 			err = err1
 		}
@@ -1566,7 +1695,7 @@ func (t *gdbThread) reloadGAtPC() error {
 
 	err = t.p.conn.step(t.strID, nil, true)
 	if err != nil {
-		if err == threadBlockedError {
+		if err == errThreadBlocked {
 			t.regs.tls = 0
 			t.regs.gaddr = 0
 			t.regs.hasgaddr = true
@@ -1575,7 +1704,7 @@ func (t *gdbThread) reloadGAtPC() error {
 		return err
 	}
 
-	if err := t.readSomeRegisters(regnamePC, regnameCX); err != nil {
+	if err := t.readSomeRegisters(t.p.regnames.PC, t.p.regnames.CX); err != nil {
 		return err
 	}
 
@@ -1602,7 +1731,7 @@ func (t *gdbThread) reloadGAlloc() error {
 	pc := t.regs.PC()
 
 	t.regs.setPC(t.p.loadGInstrAddr)
-	if err := t.writeSomeRegisters(regnamePC); err != nil {
+	if err := t.writeSomeRegisters(t.p.regnames.PC); err != nil {
 		return err
 	}
 
@@ -1611,7 +1740,7 @@ func (t *gdbThread) reloadGAlloc() error {
 	defer func() {
 		t.regs.setPC(pc)
 		t.regs.setCX(cx)
-		err1 := t.writeSomeRegisters(regnamePC, regnameCX)
+		err1 := t.writeSomeRegisters(t.p.regnames.PC, t.p.regnames.CX)
 		if err == nil {
 			err = err1
 		}
@@ -1619,7 +1748,7 @@ func (t *gdbThread) reloadGAlloc() error {
 
 	err = t.p.conn.step(t.strID, nil, true)
 	if err != nil {
-		if err == threadBlockedError {
+		if err == errThreadBlocked {
 			t.regs.tls = 0
 			t.regs.gaddr = 0
 			t.regs.hasgaddr = true
@@ -1628,7 +1757,7 @@ func (t *gdbThread) reloadGAlloc() error {
 		return err
 	}
 
-	if err := t.readSomeRegisters(regnameCX); err != nil {
+	if err := t.readSomeRegisters(t.p.regnames.CX); err != nil {
 		return err
 	}
 
@@ -1655,7 +1784,7 @@ func (t *gdbThread) SetCurrentBreakpoint(adjustPC bool) error {
 	pc := regs.PC()
 	if bp, ok := t.p.FindBreakpoint(pc); ok {
 		if t.regs.PC() != bp.Addr {
-			if err := t.SetPC(bp.Addr); err != nil {
+			if err := t.setPC(bp.Addr); err != nil {
 				return err
 			}
 		}
@@ -1671,34 +1800,27 @@ func (t *gdbThread) SetCurrentBreakpoint(adjustPC bool) error {
 }
 
 func (regs *gdbRegisters) PC() uint64 {
-	return binary.LittleEndian.Uint64(regs.regs[regnamePC].value)
+	return binary.LittleEndian.Uint64(regs.regs[regs.regnames.PC].value)
 }
 
 func (regs *gdbRegisters) setPC(value uint64) {
-	binary.LittleEndian.PutUint64(regs.regs[regnamePC].value, value)
+	binary.LittleEndian.PutUint64(regs.regs[regs.regnames.PC].value, value)
 }
 
 func (regs *gdbRegisters) SP() uint64 {
-	return binary.LittleEndian.Uint64(regs.regs[regnameSP].value)
-}
-func (regs *gdbRegisters) setSP(value uint64) {
-	binary.LittleEndian.PutUint64(regs.regs[regnameSP].value, value)
-}
-
-func (regs *gdbRegisters) setDX(value uint64) {
-	binary.LittleEndian.PutUint64(regs.regs[regnameDX].value, value)
+	return binary.LittleEndian.Uint64(regs.regs[regs.regnames.SP].value)
 }
 
 func (regs *gdbRegisters) BP() uint64 {
-	return binary.LittleEndian.Uint64(regs.regs[regnameBP].value)
+	return binary.LittleEndian.Uint64(regs.regs[regs.regnames.BP].value)
 }
 
 func (regs *gdbRegisters) CX() uint64 {
-	return binary.LittleEndian.Uint64(regs.regs[regnameCX].value)
+	return binary.LittleEndian.Uint64(regs.regs[regs.regnames.CX].value)
 }
 
 func (regs *gdbRegisters) setCX(value uint64) {
-	binary.LittleEndian.PutUint64(regs.regs[regnameCX].value, value)
+	binary.LittleEndian.PutUint64(regs.regs[regs.regnames.CX].value, value)
 }
 
 func (regs *gdbRegisters) TLS() uint64 {
@@ -1717,198 +1839,40 @@ func (regs *gdbRegisters) byName(name string) uint64 {
 	return binary.LittleEndian.Uint64(reg.value)
 }
 
-func (regs *gdbRegisters) Get(n int) (uint64, error) {
-	reg := x86asm.Reg(n)
-	const (
-		mask8  = 0x000f
-		mask16 = 0x00ff
-		mask32 = 0xffff
-	)
-
-	switch reg {
-	// 8-bit
-	case x86asm.AL:
-		return regs.byName("rax") & mask8, nil
-	case x86asm.CL:
-		return regs.byName("rcx") & mask8, nil
-	case x86asm.DL:
-		return regs.byName("rdx") & mask8, nil
-	case x86asm.BL:
-		return regs.byName("rbx") & mask8, nil
-	case x86asm.AH:
-		return (regs.byName("rax") >> 8) & mask8, nil
-	case x86asm.CH:
-		return (regs.byName("rcx") >> 8) & mask8, nil
-	case x86asm.DH:
-		return (regs.byName("rdx") >> 8) & mask8, nil
-	case x86asm.BH:
-		return (regs.byName("rbx") >> 8) & mask8, nil
-	case x86asm.SPB:
-		return regs.byName("rsp") & mask8, nil
-	case x86asm.BPB:
-		return regs.byName("rbp") & mask8, nil
-	case x86asm.SIB:
-		return regs.byName("rsi") & mask8, nil
-	case x86asm.DIB:
-		return regs.byName("rdi") & mask8, nil
-	case x86asm.R8B:
-		return regs.byName("r8") & mask8, nil
-	case x86asm.R9B:
-		return regs.byName("r9") & mask8, nil
-	case x86asm.R10B:
-		return regs.byName("r10") & mask8, nil
-	case x86asm.R11B:
-		return regs.byName("r11") & mask8, nil
-	case x86asm.R12B:
-		return regs.byName("r12") & mask8, nil
-	case x86asm.R13B:
-		return regs.byName("r13") & mask8, nil
-	case x86asm.R14B:
-		return regs.byName("r14") & mask8, nil
-	case x86asm.R15B:
-		return regs.byName("r15") & mask8, nil
-
-	// 16-bit
-	case x86asm.AX:
-		return regs.byName("rax") & mask16, nil
-	case x86asm.CX:
-		return regs.byName("rcx") & mask16, nil
-	case x86asm.DX:
-		return regs.byName("rdx") & mask16, nil
-	case x86asm.BX:
-		return regs.byName("rbx") & mask16, nil
-	case x86asm.SP:
-		return regs.byName("rsp") & mask16, nil
-	case x86asm.BP:
-		return regs.byName("rbp") & mask16, nil
-	case x86asm.SI:
-		return regs.byName("rsi") & mask16, nil
-	case x86asm.DI:
-		return regs.byName("rdi") & mask16, nil
-	case x86asm.R8W:
-		return regs.byName("r8") & mask16, nil
-	case x86asm.R9W:
-		return regs.byName("r9") & mask16, nil
-	case x86asm.R10W:
-		return regs.byName("r10") & mask16, nil
-	case x86asm.R11W:
-		return regs.byName("r11") & mask16, nil
-	case x86asm.R12W:
-		return regs.byName("r12") & mask16, nil
-	case x86asm.R13W:
-		return regs.byName("r13") & mask16, nil
-	case x86asm.R14W:
-		return regs.byName("r14") & mask16, nil
-	case x86asm.R15W:
-		return regs.byName("r15") & mask16, nil
-
-	// 32-bit
-	case x86asm.EAX:
-		return regs.byName("rax") & mask32, nil
-	case x86asm.ECX:
-		return regs.byName("rcx") & mask32, nil
-	case x86asm.EDX:
-		return regs.byName("rdx") & mask32, nil
-	case x86asm.EBX:
-		return regs.byName("rbx") & mask32, nil
-	case x86asm.ESP:
-		return regs.byName("rsp") & mask32, nil
-	case x86asm.EBP:
-		return regs.byName("rbp") & mask32, nil
-	case x86asm.ESI:
-		return regs.byName("rsi") & mask32, nil
-	case x86asm.EDI:
-		return regs.byName("rdi") & mask32, nil
-	case x86asm.R8L:
-		return regs.byName("r8") & mask32, nil
-	case x86asm.R9L:
-		return regs.byName("r9") & mask32, nil
-	case x86asm.R10L:
-		return regs.byName("r10") & mask32, nil
-	case x86asm.R11L:
-		return regs.byName("r11") & mask32, nil
-	case x86asm.R12L:
-		return regs.byName("r12") & mask32, nil
-	case x86asm.R13L:
-		return regs.byName("r13") & mask32, nil
-	case x86asm.R14L:
-		return regs.byName("r14") & mask32, nil
-	case x86asm.R15L:
-		return regs.byName("r15") & mask32, nil
-
-	// 64-bit
-	case x86asm.RAX:
-		return regs.byName("rax"), nil
-	case x86asm.RCX:
-		return regs.byName("rcx"), nil
-	case x86asm.RDX:
-		return regs.byName("rdx"), nil
-	case x86asm.RBX:
-		return regs.byName("rbx"), nil
-	case x86asm.RSP:
-		return regs.byName("rsp"), nil
-	case x86asm.RBP:
-		return regs.byName("rbp"), nil
-	case x86asm.RSI:
-		return regs.byName("rsi"), nil
-	case x86asm.RDI:
-		return regs.byName("rdi"), nil
-	case x86asm.R8:
-		return regs.byName("r8"), nil
-	case x86asm.R9:
-		return regs.byName("r9"), nil
-	case x86asm.R10:
-		return regs.byName("r10"), nil
-	case x86asm.R11:
-		return regs.byName("r11"), nil
-	case x86asm.R12:
-		return regs.byName("r12"), nil
-	case x86asm.R13:
-		return regs.byName("r13"), nil
-	case x86asm.R14:
-		return regs.byName("r14"), nil
-	case x86asm.R15:
-		return regs.byName("r15"), nil
-	}
-
-	return 0, proc.ErrUnknownRegister
-}
-
 func (r *gdbRegisters) FloatLoadError() error {
 	return nil
 }
 
 // SetPC will set the value of the PC register to the given value.
-func (t *gdbThread) SetPC(pc uint64) error {
+func (t *gdbThread) setPC(pc uint64) error {
 	_, _ = t.Registers() // Registes must be loaded first
 	t.regs.setPC(pc)
 	if t.p.gcmdok {
 		return t.p.conn.writeRegisters(t.strID, t.regs.buf)
 	}
-	reg := t.regs.regs[regnamePC]
+	reg := t.regs.regs[t.regs.regnames.PC]
 	return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
 }
 
-// SetSP will set the value of the SP register to the given value.
-func (t *gdbThread) SetSP(sp uint64) error {
-	_, _ = t.Registers() // Registes must be loaded first
-	t.regs.setSP(sp)
-	if t.p.gcmdok {
-		return t.p.conn.writeRegisters(t.strID, t.regs.buf)
+// SetReg will change the value of a list of registers
+func (t *gdbThread) SetReg(regNum uint64, reg *op.DwarfRegister) error {
+	regName := registerName(t.p.bi.Arch, regNum)
+	_, _ = t.Registers() // Registers must be loaded first
+	gdbreg, ok := t.regs.regs[regName]
+	if !ok && strings.HasPrefix(regName, "xmm") {
+		// XMMn and YMMn are the same amd64 register (in different sizes), if we
+		// don't find XMMn try YMMn instead.
+		gdbreg, ok = t.regs.regs["y"+regName[1:]]
 	}
-	reg := t.regs.regs[regnameSP]
-	return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
-}
-
-// SetDX will set the value of the DX register to the given value.
-func (t *gdbThread) SetDX(dx uint64) error {
-	_, _ = t.Registers() // Registes must be loaded first
-	t.regs.setDX(dx)
-	if t.p.gcmdok {
-		return t.p.conn.writeRegisters(t.strID, t.regs.buf)
+	if !ok {
+		return fmt.Errorf("could not set register %s: not found", regName)
 	}
-	reg := t.regs.regs[regnameDX]
-	return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
+	reg.FillBytes()
+	if len(reg.Bytes) != len(gdbreg.value) {
+		return fmt.Errorf("could not set register %s: wrong size, expected %d got %d", regName, len(gdbreg.value), len(reg.Bytes))
+	}
+	copy(gdbreg.value, reg.Bytes)
+	return t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
 }
 
 func (regs *gdbRegisters) Slice(floatingPoint bool) ([]proc.Register, error) {
@@ -1966,7 +1930,12 @@ func (regs *gdbRegisters) Slice(floatingPoint bool) ([]proc.Register, error) {
 
 func (regs *gdbRegisters) Copy() (proc.Registers, error) {
 	savedRegs := &gdbRegisters{}
-	savedRegs.init(regs.regsInfo)
+	savedRegs.init(regs.regsInfo, regs.arch, regs.regnames)
 	copy(savedRegs.buf, regs.buf)
 	return savedRegs, nil
 }
+
+func registerName(arch *proc.Arch, regNum uint64) string {
+	regName, _, _ := arch.DwarfRegisterToString(int(regNum), nil)
+	return strings.ToLower(regName)
+}
