@@ -44,20 +44,11 @@ type gdbConn struct {
 	maxTransmitAttempts   int  // maximum number of transmit or receive attempts when bad checksums are read
 	threadSuffixSupported bool // thread suffix supported by stub
 	isDebugserver         bool // true if the stub is debugserver
+	xcmdok                bool // x command can be used to transfer memory
 
 	log *logrus.Entry
 }
 
-const (
-	regnamePC     = "rip"
-	regnameCX     = "rcx"
-	regnameSP     = "rsp"
-	regnameDX     = "rdx"
-	regnameBP     = "rbp"
-	regnameFsBase = "fs_base"
-	regnameGsBase = "gs_base"
-)
-
 var ErrTooManyAttempts = errors.New("too many transmit attempts")
 
 // GdbProtocolError is an error response (Exx) of Gdb Remote Serial Protocol
@@ -103,7 +94,7 @@ const (
 	qSupportedMultiprocess = "$qSupported:multiprocess+;swbreak+;hwbreak+;no-resumed+;xmlRegisters=i386"
 )
 
-func (conn *gdbConn) handshake() error {
+func (conn *gdbConn) handshake(regnames *gdbRegnames) error {
 	conn.ack = true
 	conn.packetSize = 256
 	conn.rdr = bufio.NewReader(conn.conn)
@@ -149,15 +140,29 @@ func (conn *gdbConn) handshake() error {
 
 	// Attempt to figure out the name of the processor register.
 	// We either need qXfer:features:read (gdbserver/rr) or qRegisterInfo (lldb)
-	if err := conn.readRegisterInfo(); err != nil {
+	regFound := map[string]bool{
+		regnames.PC: false,
+		regnames.SP: false,
+		regnames.BP: false,
+		regnames.CX: false,
+	}
+	if err := conn.readRegisterInfo(regFound); err != nil {
 		if isProtocolErrorUnsupported(err) {
-			if err := conn.readTargetXml(); err != nil {
+			if err := conn.readTargetXml(regFound); err != nil {
 				return err
 			}
 		} else {
 			return err
 		}
 	}
+	for n := range regFound {
+		if n == "" {
+			continue
+		}
+		if !regFound[n] {
+			return fmt.Errorf("could not find %s register", n)
+		}
+	}
 
 	// We either need:
 	//  * QListThreadsInStopReply + qThreadStopInfo (i.e. lldb-server/debugserver),
@@ -173,6 +178,10 @@ func (conn *gdbConn) handshake() error {
 		}
 	}
 
+	if resp, err := conn.exec([]byte("$x0,0"), "init"); err == nil && string(resp) == "OK" {
+		conn.xcmdok = true
+	}
+
 	return nil
 }
 
@@ -231,17 +240,24 @@ type gdbRegisterInfo struct {
 	Group   string `xml:"group,attr"`
 }
 
+func setRegFound(regFound map[string]bool, name string) {
+	for n := range regFound {
+		if name == n {
+			regFound[n] = true
+		}
+	}
+}
+
 // readTargetXml reads target.xml file from stub using qXfer:features:read,
 // then parses it requesting any additional files.
 // The schema of target.xml is described by:
 //  https://github.com/bminor/binutils-gdb/blob/61baf725eca99af2569262d10aca03dcde2698f6/gdb/features/gdb-target.dtd
-func (conn *gdbConn) readTargetXml() (err error) {
+func (conn *gdbConn) readTargetXml(regFound map[string]bool) (err error) {
 	conn.regsInfo, err = conn.readAnnex("target.xml")
 	if err != nil {
 		return err
 	}
 	var offset int
-	var pcFound, cxFound, spFound bool
 	regnum := 0
 	for i := range conn.regsInfo {
 		if conn.regsInfo[i].Regnum == 0 {
@@ -251,33 +267,18 @@ func (conn *gdbConn) readTargetXml() (err error) {
 		}
 		conn.regsInfo[i].Offset = offset
 		offset += conn.regsInfo[i].Bitsize / 8
-		switch conn.regsInfo[i].Name {
-		case regnamePC:
-			pcFound = true
-		case regnameCX:
-			cxFound = true
-		case regnameSP:
-			spFound = true
-		}
+
+		setRegFound(regFound, conn.regsInfo[i].Name)
 		regnum++
 	}
-	if !pcFound {
-		return errors.New("could not find RIP register")
-	}
-	if !spFound {
-		return errors.New("could not find RSP register")
-	}
-	if !cxFound {
-		return errors.New("could not find RCX register")
-	}
+
 	return nil
 }
 
 // readRegisterInfo uses qRegisterInfo to read register information (used
 // when qXfer:feature:read is not supported).
-func (conn *gdbConn) readRegisterInfo() (err error) {
+func (conn *gdbConn) readRegisterInfo(regFound map[string]bool) (err error) {
 	regnum := 0
-	var pcFound, cxFound, spFound bool
 	for {
 		conn.outbuf.Reset()
 		fmt.Fprintf(&conn.outbuf, "$qRegisterInfo%x", regnum)
@@ -330,30 +331,13 @@ func (conn *gdbConn) readRegisterInfo() (err error) {
 			continue
 		}
 
-		switch regname {
-		case regnamePC:
-			pcFound = true
-		case regnameCX:
-			cxFound = true
-		case regnameSP:
-			spFound = true
-		}
+		setRegFound(regFound, regname)
 
 		conn.regsInfo = append(conn.regsInfo, gdbRegisterInfo{Regnum: regnum, Name: regname, Bitsize: bitsize, Offset: offset})
 
 		regnum++
 	}
 
-	if !pcFound {
-		return errors.New("could not find RIP register")
-	}
-	if !spFound {
-		return errors.New("could not find RSP register")
-	}
-	if !cxFound {
-		return errors.New("could not find RCX register")
-	}
-
 	return nil
 }
 
@@ -412,18 +396,18 @@ func (conn *gdbConn) qXfer(kind, annex string, binary bool) ([]byte, error) {
 	return out, nil
 }
 
-// setBreakpoint executes a 'Z' (insert breakpoint) command of type '0' and kind '1'
-func (conn *gdbConn) setBreakpoint(addr uint64) error {
+// setBreakpoint executes a 'Z' (insert breakpoint) command of type '0' and kind '1' or '4'
+func (conn *gdbConn) setBreakpoint(addr uint64, kind int) error {
 	conn.outbuf.Reset()
-	fmt.Fprintf(&conn.outbuf, "$Z0,%x,1", addr)
+	fmt.Fprintf(&conn.outbuf, "$Z0,%x,%d", addr, kind)
 	_, err := conn.exec(conn.outbuf.Bytes(), "set breakpoint")
 	return err
 }
 
-// clearBreakpoint executes a 'z' (remove breakpoint) command of type '0' and kind '1'
-func (conn *gdbConn) clearBreakpoint(addr uint64) error {
+// clearBreakpoint executes a 'z' (remove breakpoint) command of type '0' and kind '1' or '4'
+func (conn *gdbConn) clearBreakpoint(addr uint64, kind int) error {
 	conn.outbuf.Reset()
-	fmt.Fprintf(&conn.outbuf, "$z0,%x,1", addr)
+	fmt.Fprintf(&conn.outbuf, "$z0,%x,%d", addr, kind)
 	_, err := conn.exec(conn.outbuf.Bytes(), "clear breakpoint")
 	return err
 }
@@ -632,7 +616,7 @@ func (conn *gdbConn) step(threadID string, tu *threadUpdater, ignoreFaultSignal
 	}
 }
 
-var threadBlockedError = errors.New("thread blocked")
+var errThreadBlocked = errors.New("thread blocked")
 
 func (conn *gdbConn) waitForvContStop(context string, threadID string, tu *threadUpdater) (string, uint8, error) {
 	count := 0
@@ -655,7 +639,7 @@ func (conn *gdbConn) waitForvContStop(context string, threadID string, tu *threa
 			}
 			count++
 		} else if failed {
-			return "", 0, threadBlockedError
+			return "", 0, errThreadBlocked
 		} else if err != nil {
 			return "", 0, err
 		} else {
@@ -879,8 +863,15 @@ func (conn *gdbConn) appendThreadSelector(threadID string) {
 	fmt.Fprintf(&conn.outbuf, ";thread:%s;", threadID)
 }
 
-// executes 'm' (read memory) command
 func (conn *gdbConn) readMemory(data []byte, addr uint64) error {
+	if conn.xcmdok && len(data) > conn.packetSize {
+		return conn.readMemoryBinary(data, addr)
+	}
+	return conn.readMemoryHex(data, addr)
+}
+
+// executes 'm' (read memory) command
+func (conn *gdbConn) readMemoryHex(data []byte, addr uint64) error {
 	size := len(data)
 	data = data[:0]
 
@@ -908,6 +899,29 @@ func (conn *gdbConn) readMemory(data []byte, addr uint64) error {
 	return nil
 }
 
+// executes 'x' (binary read memory) command
+func (conn *gdbConn) readMemoryBinary(data []byte, addr uint64) error {
+	size := len(data)
+	data = data[:0]
+
+	for len(data) < size {
+		conn.outbuf.Reset()
+
+		sz := size - len(data)
+
+		fmt.Fprintf(&conn.outbuf, "$x%x,%x", addr+uint64(len(data)), sz)
+		if err := conn.send(conn.outbuf.Bytes()); err != nil {
+			return err
+		}
+		resp, err := conn.recv(conn.outbuf.Bytes(), "binary memory read", true)
+		if err != nil {
+			return err
+		}
+		data = append(data, resp...)
+	}
+	return nil
+}
+
 func writeAsciiBytes(w io.Writer, data []byte) {
 	for _, b := range data {
 		fmt.Fprintf(w, "%02x", b)
@@ -999,8 +1013,9 @@ type imageList struct {
 }
 
 type imageDescription struct {
-	Pathname   string     `json:"pathname"`
-	MachHeader machHeader `json:"mach_header"`
+	LoadAddress uint64     `json:"load_address"`
+	Pathname    string     `json:"pathname"`
+	MachHeader  machHeader `json:"mach_header"`
 }
 
 type machHeader struct {
@@ -1023,6 +1038,87 @@ func (conn *gdbConn) getLoadedDynamicLibraries() ([]imageDescription, error) {
 	return images.Images, err
 }
 
+type memoryRegionInfo struct {
+	start       uint64
+	size        uint64
+	permissions string
+	name        string
+}
+
+func decodeHexString(in []byte) (string, bool) {
+	out := make([]byte, 0, len(in)/2)
+	for i := 0; i < len(in); i += 2 {
+		v, err := strconv.ParseUint(string(in[i:i+2]), 16, 8)
+		if err != nil {
+			return "", false
+		}
+		out = append(out, byte(v))
+	}
+	return string(out), true
+}
+
+func (conn *gdbConn) memoryRegionInfo(addr uint64) (*memoryRegionInfo, error) {
+	conn.outbuf.Reset()
+	fmt.Fprintf(&conn.outbuf, "$qMemoryRegionInfo:%x", addr)
+	resp, err := conn.exec(conn.outbuf.Bytes(), "qMemoryRegionInfo")
+	if err != nil {
+		return nil, err
+	}
+
+	mri := &memoryRegionInfo{}
+
+	buf := resp
+	for len(buf) > 0 {
+		colon := bytes.Index(buf, []byte{':'})
+		if colon < 0 {
+			break
+		}
+		key := buf[:colon]
+		buf = buf[colon+1:]
+
+		semicolon := bytes.Index(buf, []byte{';'})
+		var value []byte
+		if semicolon < 0 {
+			value = buf
+			buf = nil
+		} else {
+			value = buf[:semicolon]
+			buf = buf[semicolon+1:]
+		}
+
+		switch string(key) {
+		case "start":
+			start, err := strconv.ParseUint(string(value), 16, 64)
+			if err != nil {
+				return nil, fmt.Errorf("malformed qMemoryRegionInfo response packet (start): %v in %s", err, string(resp))
+			}
+			mri.start = start
+		case "size":
+			size, err := strconv.ParseUint(string(value), 16, 64)
+			if err != nil {
+				return nil, fmt.Errorf("malformed qMemoryRegionInfo response packet (size): %v in %s", err, string(resp))
+			}
+			mri.size = size
+		case "permissions":
+			mri.permissions = string(value)
+		case "name":
+			namestr, ok := decodeHexString(value)
+			if !ok {
+				return nil, fmt.Errorf("malformed qMemoryRegionInfo response packet (name): %s", string(resp))
+			}
+			mri.name = namestr
+		case "error":
+			errstr, ok := decodeHexString(value)
+			if !ok {
+				return nil, fmt.Errorf("malformed qMemoryRegionInfo response packet (error): %s", string(resp))
+			}
+			return nil, fmt.Errorf("qMemoryRegionInfo error: %s", errstr)
+		}
+	}
+
+	return mri, nil
+}
+
 // exec executes a message to the stub and reads a response.
 // The details of the wire protocol are described here:
 //  https://sourceware.org/gdb/onlinedocs/gdb/Overview.html#Overview
@@ -1267,7 +1363,10 @@ func checksumok(packet, checksumBuf []byte) bool {
 	if err != nil {
 		return false
 	}
-	return sum == uint8(tgt)
+
+	tgt8 := uint8(tgt)
+
+	return sum == tgt8
 }
 
 func checksum(packet []byte) (sum uint8) {
