@@ -187,21 +187,33 @@ func initialize(dbp *nativeProcess) error {
 }
 
 // kill kills the target process.
-func (dbp *nativeProcess) kill() (err error) {
+func (dbp *nativeProcess) kill() error {
 	if dbp.exited {
 		return nil
 	}
 	if !dbp.threads[dbp.pid].Stopped() {
 		return errors.New("process must be stopped in order to kill it")
 	}
-	if err = sys.Kill(-dbp.pid, sys.SIGKILL); err != nil {
+	if err := sys.Kill(-dbp.pid, sys.SIGKILL); err != nil {
 		return errors.New("could not deliver signal " + err.Error())
 	}
-	if _, _, err = dbp.wait(dbp.pid, 0); err != nil {
-		return
+	// wait for other threads first or the thread group leader (dbp.pid) will never exit.
+	for threadID := range dbp.threads {
+		if threadID != dbp.pid {
+			dbp.wait(threadID, 0)
+		}
 	}
-	dbp.postExit()
-	return
+	for {
+		wpid, status, err := dbp.wait(dbp.pid, 0)
+		if err != nil {
+			return err
+		}
+		if wpid == dbp.pid && status != nil && status.Signaled() && status.Signal() == sys.SIGKILL {
+			dbp.postExit()
+			return err
+		}
+	}
+	return nil
 }
 
 func (dbp *nativeProcess) requestManualStop() (err error) {
@@ -256,6 +268,14 @@ func (dbp *nativeProcess) addThread(tid int, attach bool) (*nativeThread, error)
 	if dbp.memthread == nil {
 		dbp.memthread = dbp.threads[tid]
 	}
+	for _, bp := range dbp.Breakpoints().M {
+		if bp.WatchType != 0 {
+			err := dbp.threads[tid].writeHardwareBreakpoint(bp.Addr, bp.WatchType, bp.HWBreakIndex)
+			if err != nil {
+				return nil, err
+			}
+		}
+	}
 	return dbp.threads[tid], nil
 }
 
@@ -322,6 +342,16 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 			delete(dbp.threads, wpid)
 			continue
 		}
+		if status.Signaled() {
+			// Signaled means the thread was terminated due to a signal.
+			if wpid == dbp.pid {
+				dbp.postExit()
+				return nil, proc.ErrProcessExited{Pid: wpid, Status: -int(status.Signal())}
+			}
+			// does this ever happen?
+			delete(dbp.threads, wpid)
+			continue
+		}
 		if status.StopSignal() == sys.SIGTRAP && status.TrapCause() == sys.PTRACE_EVENT_CLONE {
 			// A traced thread has cloned a new thread, grab the pid and
 			// add it to our list of traced threads.
@@ -386,10 +416,15 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 			th.os.running = false
 			return th, nil
 		} else if err := th.resumeWithSig(int(status.StopSignal())); err != nil {
-			if options&trapWaitDontCallExitGuard != 0 {
+			if err != sys.ESRCH {
 				return nil, err
 			}
-			return nil, dbp.exitGuard(err)
+			// do the same thing we do if a thread quit
+			if wpid == dbp.pid {
+				dbp.postExit()
+				return nil, proc.ErrProcessExited{Pid: wpid, Status: status.ExitStatus()}
+			}
+			delete(dbp.threads, wpid)
 		}
 	}
 }
@@ -488,7 +523,7 @@ func (dbp *nativeProcess) resume() error {
 // stop stops all running threads and sets breakpoints
 func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error) {
 	if dbp.exited {
-		return nil, &proc.ErrProcessExited{Pid: dbp.Pid()}
+		return nil, proc.ErrProcessExited{Pid: dbp.Pid()}
 	}
 
 	for _, th := range dbp.threads {
@@ -541,12 +576,28 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 	switchTrapthread := false
 
 	// set breakpoints on SIGTRAP threads
+	var err1 error
 	for _, th := range dbp.threads {
-		if th.CurrentBreakpoint.Breakpoint == nil && th.os.setbp {
-			if err := th.SetCurrentBreakpoint(true); err != nil {
-				return nil, err
+		pc, _ := th.PC()
+
+		if !th.os.setbp && pc != th.os.phantomBreakpointPC {
+			// check if this could be a breakpoint hit anyway that the OS hasn't notified us about, yet.
+			if _, ok := dbp.FindBreakpoint(pc, dbp.BinInfo().Arch.BreakInstrMovesPC()); ok {
+				th.os.phantomBreakpointPC = pc
 			}
 		}
+
+		if pc != th.os.phantomBreakpointPC {
+			th.os.phantomBreakpointPC = 0
+		}
+
+		if th.CurrentBreakpoint.Breakpoint == nil && th.os.setbp {
+			if err := th.SetCurrentBreakpoint(true); err != nil {
+				err1 = err
+				continue
+			}
+		}
+
 		if th.CurrentBreakpoint.Breakpoint == nil && th.os.setbp && (th.Status != nil) && ((*sys.WaitStatus)(th.Status).StopSignal() == sys.SIGTRAP) && dbp.BinInfo().Arch.BreakInstrMovesPC() {
 			manualStop := false
 			if th.ThreadID() == trapthread.ThreadID() {
@@ -554,7 +605,7 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 				manualStop = dbp.manualStopRequested
 				dbp.stopMu.Unlock()
 			}
-			if !manualStop {
+			if !manualStop && th.os.phantomBreakpointPC == pc {
 				// Thread received a SIGTRAP but we don't have a breakpoint for it and
 				// it wasn't sent by a manual stop request. It's either a hardcoded
 				// breakpoint or a phantom breakpoint hit (a breakpoint that was hit but
@@ -577,7 +628,7 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 				}
 				if !isHardcodedBreakpoint {
 					// phantom breakpoint hit
-					_ = th.SetPC(pc - uint64(len(dbp.BinInfo().Arch.BreakpointInstruction())))
+					_ = th.setPC(pc - uint64(len(dbp.BinInfo().Arch.BreakpointInstruction())))
 					th.os.setbp = false
 					if trapthread.ThreadID() == th.ThreadID() {
 						// Will switch to a different thread for trapthread because we don't
@@ -589,11 +640,15 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 			}
 		}
 	}
+	if err1 != nil {
+		return nil, err1
+	}
 
 	if switchTrapthread {
+		trapthreadID := trapthread.ID
 		trapthread = nil
 		for _, th := range dbp.threads {
-			if th.os.setbp && th.ThreadID() != trapthread.ThreadID() {
+			if th.os.setbp && th.ThreadID() != trapthreadID {
 				return th, nil
 			}
 		}
