@@ -100,7 +100,9 @@ func ThreadStacktrace(thread Thread, depth int) ([]Stackframe, error) {
 			return nil, err
 		}
 		so := thread.BinInfo().PCToImage(regs.PC())
-		it := newStackIterator(thread.BinInfo(), thread.ProcessMemory(), thread.BinInfo().Arch.RegistersToDwarfRegisters(so.StaticBase, regs), 0, nil, -1, nil, 0)
+		dwarfRegs := *(thread.BinInfo().Arch.RegistersToDwarfRegisters(so.StaticBase, regs))
+		dwarfRegs.ChangeFunc = thread.SetReg
+		it := newStackIterator(thread.BinInfo(), thread.ProcessMemory(), dwarfRegs, 0, nil, -1, nil, 0)
 		return it.stacktrace(depth)
 	}
 	return g.Stacktrace(depth, 0)
@@ -119,9 +121,11 @@ func (g *G) stackIterator(opts StacktraceOptions) (*stackIterator, error) {
 			return nil, err
 		}
 		so := bi.PCToImage(regs.PC())
+		dwarfRegs := *(bi.Arch.RegistersToDwarfRegisters(so.StaticBase, regs))
+		dwarfRegs.ChangeFunc = g.Thread.SetReg
 		return newStackIterator(
 			bi, g.variable.mem,
-			bi.Arch.RegistersToDwarfRegisters(so.StaticBase, regs),
+			dwarfRegs,
 			g.stack.hi, stkbar, g.stkbarPos, g, opts), nil
 	}
 	so := g.variable.bi.PCToImage(g.PC)
@@ -377,7 +381,9 @@ func (it *stackIterator) appendInlineCalls(frames []Stackframe, frame Stackframe
 		if !okname || !okfileidx || !okline {
 			break
 		}
-		if fileidx-1 < 0 || fileidx-1 >= int64(len(frame.Current.Fn.cu.lineInfo.FileNames)) {
+		var e *dwarf.Entry
+		filepath, fileErr := frame.Current.Fn.cu.filePath(int(fileidx), e)
+		if fileErr != nil {
 			break
 		}
 
@@ -400,16 +406,22 @@ func (it *stackIterator) appendInlineCalls(frames []Stackframe, frame Stackframe
 			lastpc:      frame.lastpc,
 		})
 
-		frame.Call.File = frame.Current.Fn.cu.lineInfo.FileNames[fileidx-1].Path
+		frame.Call.File = filepath
 		frame.Call.Line = int(line)
 	}
 
 	return append(frames, frame)
 }
 
-// advanceRegs calculates it.callFrameRegs using it.regs and the frame
-// descriptor entry for the current stack frame.
-// it.regs.CallFrameCFA is updated.
+// advanceRegs calculates the DwarfRegisters for a next stack frame
+// (corresponding to it.pc).
+//
+// The computation uses the registers for the current stack frame (it.regs) and
+// the corresponding Frame Descriptor Entry (FDE) retrieved from the DWARF info.
+//
+// The new set of registers is returned. it.regs is not updated, except for
+// it.regs.CFA; the caller has to eventually switch it.regs when the iterator
+// advances to the next frame.
 func (it *stackIterator) advanceRegs() (callFrameRegs op.DwarfRegisters, ret uint64, retaddr uint64) {
 	fde, err := it.bi.frameEntries.FDEForPC(it.pc)
 	var framectx *frame.FrameContext
