@@ -20,9 +20,11 @@ import (
 	"strconv"
 	"strings"
 	"text/tabwriter"
+	"time"
 
 	"github.com/cosiner/argv"
 	"github.com/go-delve/delve/pkg/locspec"
+	"github.com/go-delve/delve/pkg/terminal/colorize"
 	"github.com/go-delve/delve/service"
 	"github.com/go-delve/delve/service/api"
 	"github.com/go-delve/delve/service/rpc2"
@@ -132,7 +134,7 @@ See also: "help on", "help cond" and "help clear"`},
 
 For recorded targets the command takes the following forms:
 
-	restart					resets ot the start of the recording
+	restart					resets to the start of the recording
 	restart [checkpoint]			resets the recording to the given checkpoint
 	restart -r [newargv...]	[redirects...]	re-records the target process
 	
@@ -150,12 +152,22 @@ A list of file redirections can be specified after the new argument list to over
 	2>error.txt	redirects the standard error of the target process to error.txt
 `},
 		{aliases: []string{"rebuild"}, group: runCmds, cmdFn: c.rebuild, allowedPrefixes: revPrefix, helpMsg: "Rebuild the target executable and restarts it. It does not work if the executable was not built by delve."},
-		{aliases: []string{"continue", "c"}, group: runCmds, cmdFn: c.cont, allowedPrefixes: revPrefix, helpMsg: "Run until breakpoint or program termination."},
+		{aliases: []string{"continue", "c"}, group: runCmds, cmdFn: c.cont, allowedPrefixes: revPrefix, helpMsg: `Run until breakpoint or program termination.
+
+	continue [<linespec>]
+
+Optional linespec argument allows you to continue until a specific location is reached. The program will halt if a breakpoint is hit before reaching the specified location.
+
+For example:
+
+	continue main.main
+	continue encoding/json.Marshal
+`},
 		{aliases: []string{"step", "s"}, group: runCmds, cmdFn: c.step, allowedPrefixes: revPrefix, helpMsg: "Single step through program."},
 		{aliases: []string{"step-instruction", "si"}, group: runCmds, allowedPrefixes: revPrefix, cmdFn: c.stepInstruction, helpMsg: "Single step a single cpu instruction."},
 		{aliases: []string{"next", "n"}, group: runCmds, cmdFn: c.next, allowedPrefixes: revPrefix, helpMsg: `Step over to next source line.
 
-	 next [count]
+	next [count]
 
 Optional [count] argument allows you to skip multiple lines.
 `},
@@ -188,6 +200,9 @@ Current limitations:
 	clearall [<linespec>]
 
 If called with the linespec argument it will delete all the breakpoints matching the linespec. If linespec is omitted all breakpoints are deleted.`},
+		{aliases: []string{"toggle"}, group: breakCmds, cmdFn: toggle, helpMsg: `Toggles on or off a breakpoint.
+
+toggle <breakpoint name or id>`},
 		{aliases: []string{"goroutines", "grs"}, group: goroutineCmds, cmdFn: goroutines, helpMsg: `List program goroutines.
 
 	goroutines [-u (default: user location)|-r (runtime location)|-g (go statement location)|-s (start location)] [-t (stack trace)] [-l (labels)]
@@ -214,9 +229,11 @@ Called with more arguments it will execute a command on the specified goroutine.
 		{aliases: []string{"breakpoints", "bp"}, group: breakCmds, cmdFn: breakpoints, helpMsg: "Print out info for active breakpoints."},
 		{aliases: []string{"print", "p"}, group: dataCmds, allowedPrefixes: onPrefix | deferredPrefix, cmdFn: printVar, helpMsg: `Evaluate an expression.
 
-	[goroutine <n>] [frame <m>] print <expression>
+	[goroutine <n>] [frame <m>] print [%format] <expression>
 
-See $GOPATH/src/github.com/go-delve/delve/Documentation/cli/expr.md for a description of supported expressions.`},
+See $GOPATH/src/github.com/go-delve/delve/Documentation/cli/expr.md for a description of supported expressions.
+
+The optional format argument is a format specifier, like the ones used by the fmt package. For example "print %x v" will print v as an hexadecimal number.`},
 		{aliases: []string{"whatis"}, group: dataCmds, cmdFn: whatisCommand, helpMsg: `Prints type of an expression.
 
 	whatis <expression>`},
@@ -261,7 +278,7 @@ If regex is specified only package variables with a name matching it will be ret
 
 	regs [-a]
 
-Argument -a shows more registers.`},
+Argument -a shows more registers. Individual registers can also be displayed by 'print' and 'display'. See $GOPATH/src/github.com/go-delve/delve/Documentation/cli/expr.md.`},
 		{aliases: []string{"exit", "quit", "q"}, cmdFn: exitCommand, helpMsg: `Exit the debugger.
 		
 	exit [-c]
@@ -391,23 +408,34 @@ If locspec is omitted edit will open the current source file in the editor, othe
 		{aliases: []string{"examinemem", "x"}, group: dataCmds, cmdFn: examineMemoryCmd, helpMsg: `Examine memory:
 
 	examinemem [-fmt <format>] [-count|-len <count>] [-size <size>] <address>
+	examinemem [-fmt <format>] [-count|-len <count>] [-size <size>] -x <expression>
 
 Format represents the data format and the value is one of this list (default hex): bin(binary), oct(octal), dec(decimal), hex(hexadecimal), addr(address).
 Length is the number of bytes (default 1) and must be less than or equal to 1000.
 Address is the memory location of the target to examine. Please note '-len' is deprecated by '-count and -size'.
+Expression can be an integer expression or pointer value of the memory location to examine.
 
 For example:
 
-    x -fmt hex -count 20 -size 1 0xc00008af38`},
+    x -fmt hex -count 20 -size 1 0xc00008af38
+    x -fmt hex -count 20 -size 1 -x 0xc00008af38 + 8
+    x -fmt hex -count 20 -size 1 -x &myVar
+    x -fmt hex -count 20 -size 1 -x myPtrVar`},
 
 		{aliases: []string{"display"}, group: dataCmds, cmdFn: display, helpMsg: `Print value of an expression every time the program stops.
 
-	display -a <expression>
+	display -a [%format] <expression>
 	display -d <number>
 
 The '-a' option adds an expression to the list of expression printed every time the program stops. The '-d' option removes the specified expression from the list.
 
 If display is called without arguments it will print the value of all expression in the list.`},
+
+		{aliases: []string{"dump"}, cmdFn: dump, helpMsg: `Creates a core dump from the current process state
+
+	dump <output file>
+
+The core dump is always written in ELF, even on systems (windows, macOS) where this is not customary. For environments other than linux/amd64 threads and registers are dumped in a format that only Delve can read back.`},
 	}
 
 	addrecorded := client == nil
@@ -723,7 +751,12 @@ func goroutines(t *Term, ctx callContext, argstr string) error {
 		gslen = 0
 		gs    []*api.Goroutine
 	)
+	t.longCommandStart()
 	for start >= 0 {
+		if t.longCommandCanceled() {
+			fmt.Printf("interrupted\n")
+			return nil
+		}
 		gs, start, err = t.client.ListGoroutines(start, goroutineBatchSize)
 		if err != nil {
 			return err
@@ -917,11 +950,58 @@ func (t *Term) formatGoroutine(g *api.Goroutine, fgl formatGoroutineLoc) string
 		locname = "Start"
 		loc = g.StartLoc
 	}
-	thread := ""
+
+	buf := new(strings.Builder)
+	fmt.Fprintf(buf, "%d - %s: %s", g.ID, locname, t.formatLocation(loc))
 	if g.ThreadID != 0 {
-		thread = fmt.Sprintf(" (thread %d)", g.ThreadID)
+		fmt.Fprintf(buf, " (thread %d)", g.ThreadID)
 	}
-	return fmt.Sprintf("%d - %s: %s%s", g.ID, locname, t.formatLocation(loc), thread)
+
+	if (g.Status == api.GoroutineWaiting || g.Status == api.GoroutineSyscall) && g.WaitReason != 0 {
+		var wr string
+		if g.WaitReason > 0 && g.WaitReason < int64(len(waitReasonStrings)) {
+			wr = waitReasonStrings[g.WaitReason]
+		} else {
+			wr = fmt.Sprintf("unknown wait reason %d", g.WaitReason)
+		}
+		fmt.Fprintf(buf, " [%s", wr)
+		if g.WaitSince > 0 {
+			fmt.Fprintf(buf, " %s", time.Since(time.Unix(0, g.WaitSince)).String())
+		}
+		fmt.Fprintf(buf, "]")
+	}
+
+	return buf.String()
+}
+
+var waitReasonStrings = [...]string{
+	"",
+	"GC assist marking",
+	"IO wait",
+	"chan receive (nil chan)",
+	"chan send (nil chan)",
+	"dumping heap",
+	"garbage collection",
+	"garbage collection scan",
+	"panicwait",
+	"select",
+	"select (no cases)",
+	"GC assist wait",
+	"GC sweep wait",
+	"GC scavenge wait",
+	"chan receive",
+	"chan send",
+	"finalizer wait",
+	"force gc (idle)",
+	"semacquire",
+	"sleep",
+	"sync.Cond.Wait",
+	"timer goroutine (idle)",
+	"trace reader (blocked)",
+	"wait for GC cycle",
+	"GC worker (idle)",
+	"preempted",
+	"debug call",
 }
 
 func writeGoroutineLong(t *Term, w io.Writer, g *api.Goroutine, prefix string) {
@@ -1131,6 +1211,19 @@ func (c *Commands) rebuild(t *Term, ctx callContext, args string) error {
 }
 
 func (c *Commands) cont(t *Term, ctx callContext, args string) error {
+	if args != "" {
+		tmp, err := setBreakpoint(t, ctx, false, args)
+		if err != nil {
+			return err
+		}
+		defer func() {
+			for _, bp := range tmp {
+				if _, err := t.client.ClearBreakpoint(bp.ID); err != nil {
+					fmt.Printf("failed to clear temporary breakpoint: %d", bp.ID)
+				}
+			}
+		}()
+	}
 	if ctx.Prefix == revPrefix {
 		return c.rewind(t, ctx, args)
 	}
@@ -1387,6 +1480,24 @@ func clearAll(t *Term, ctx callContext, args string) error {
 	return nil
 }
 
+func toggle(t *Term, ctx callContext, args string) error {
+	if args == "" {
+		return fmt.Errorf("not enough arguments")
+	}
+	id, err := strconv.Atoi(args)
+	var bp *api.Breakpoint
+	if err == nil {
+		bp, err = t.client.ToggleBreakpoint(id)
+	} else {
+		bp, err = t.client.ToggleBreakpointByName(args)
+	}
+	if err != nil {
+		return err
+	}
+	fmt.Printf("%s toggled at %s\n", formatBreakpointName(bp, true), t.formatBreakpointLocation(bp))
+	return nil
+}
+
 // byID sorts breakpoints by ID.
 type byID []*api.Breakpoint
 
@@ -1437,7 +1548,7 @@ func breakpoints(t *Term, ctx callContext, args string) error {
 	return nil
 }
 
-func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) error {
+func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) ([]*api.Breakpoint, error) {
 	args := split2PartsBySpace(argstr)
 
 	requestedBp := &api.Breakpoint{}
@@ -1453,23 +1564,24 @@ func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) err
 			spec = argstr
 		}
 	default:
-		return fmt.Errorf("address required")
+		return nil, fmt.Errorf("address required")
 	}
 
 	requestedBp.Tracepoint = tracepoint
 	locs, err := t.client.FindLocation(ctx.Scope, spec, true, t.substitutePathRules())
 	if err != nil {
 		if requestedBp.Name == "" {
-			return err
+			return nil, err
 		}
 		requestedBp.Name = ""
 		spec = argstr
 		var err2 error
 		locs, err2 = t.client.FindLocation(ctx.Scope, spec, true, t.substitutePathRules())
 		if err2 != nil {
-			return err
+			return nil, err
 		}
 	}
+	created := []*api.Breakpoint{}
 	for _, loc := range locs {
 		requestedBp.Addr = loc.PC
 		requestedBp.Addrs = loc.PCs
@@ -1479,8 +1591,9 @@ func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) err
 
 		bp, err := t.client.CreateBreakpoint(requestedBp)
 		if err != nil {
-			return err
+			return nil, err
 		}
+		created = append(created, bp)
 
 		fmt.Printf("%s set at %s\n", formatBreakpointName(bp, true), t.formatBreakpointLocation(bp))
 	}
@@ -1488,7 +1601,7 @@ func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) err
 	var shouldSetReturnBreakpoints bool
 	loc, err := locspec.Parse(spec)
 	if err != nil {
-		return err
+		return nil, err
 	}
 	switch t := loc.(type) {
 	case *locspec.NormalLocationSpec:
@@ -1503,7 +1616,7 @@ func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) err
 			}
 			addrs, err := t.client.(*rpc2.RPCClient).FunctionReturnLocations(locs[0].Function.Name())
 			if err != nil {
-				return err
+				return nil, err
 			}
 			for j := range addrs {
 				_, err = t.client.CreateBreakpoint(&api.Breakpoint{
@@ -1513,20 +1626,22 @@ func setBreakpoint(t *Term, ctx callContext, tracepoint bool, argstr string) err
 					LoadArgs:    &ShortLoadConfig,
 				})
 				if err != nil {
-					return err
+					return nil, err
 				}
 			}
 		}
 	}
-	return nil
+	return created, nil
 }
 
 func breakpoint(t *Term, ctx callContext, args string) error {
-	return setBreakpoint(t, ctx, false, args)
+	_, err := setBreakpoint(t, ctx, false, args)
+	return err
 }
 
 func tracepoint(t *Term, ctx callContext, args string) error {
-	return setBreakpoint(t, ctx, true, args)
+	_, err := setBreakpoint(t, ctx, true, args)
+	return err
 }
 
 func edit(t *Term, ctx callContext, args string) error {
@@ -1549,27 +1664,42 @@ func edit(t *Term, ctx callContext, args string) error {
 	return cmd.Run()
 }
 
-func examineMemoryCmd(t *Term, ctx callContext, args string) error {
-	v := strings.FieldsFunc(args, func(c rune) bool {
-		return c == ' '
-	})
-
+func examineMemoryCmd(t *Term, ctx callContext, argstr string) error {
 	var (
 		address uint64
 		err     error
 		ok      bool
+		args    = strings.Split(argstr, " ")
 	)
 
 	// Default value
 	priFmt := byte('x')
 	count := 1
 	size := 1
+	isExpr := false
 
-	for i := 0; i < len(v); i++ {
-		switch v[i] {
+	// nextArg returns the next argument that is not an empty string, if any, and
+	// advances the args slice to the position after that.
+	nextArg := func() string {
+		for len(args) > 0 {
+			arg := args[0]
+			args = args[1:]
+			if arg != "" {
+				return arg
+			}
+		}
+		return ""
+	}
+
+loop:
+	for {
+		switch cmd := nextArg(); cmd {
+		case "":
+			// no more arguments
+			break loop
 		case "-fmt":
-			i++
-			if i >= len(v) {
+			arg := nextArg()
+			if arg == "" {
 				return fmt.Errorf("expected argument after -fmt")
 			}
 			fmtMapToPriFmt := map[string]byte{
@@ -1582,39 +1712,39 @@ func examineMemoryCmd(t *Term, ctx callContext, args string) error {
 				"bin":         'b',
 				"binary":      'b',
 			}
-			priFmt, ok = fmtMapToPriFmt[v[i]]
+			priFmt, ok = fmtMapToPriFmt[arg]
 			if !ok {
-				return fmt.Errorf("%q is not a valid format", v[i])
+				return fmt.Errorf("%q is not a valid format", arg)
 			}
 		case "-count", "-len":
-			i++
-			if i >= len(v) {
+			arg := nextArg()
+			if arg == "" {
 				return fmt.Errorf("expected argument after -count/-len")
 			}
 			var err error
-			count, err = strconv.Atoi(v[i])
+			count, err = strconv.Atoi(arg)
 			if err != nil || count <= 0 {
 				return fmt.Errorf("count/len must be a positive integer")
 			}
 		case "-size":
-			i++
-			if i >= len(v) {
+			arg := nextArg()
+			if arg == "" {
 				return fmt.Errorf("expected argument after -size")
 			}
 			var err error
-			size, err = strconv.Atoi(v[i])
+			size, err = strconv.Atoi(arg)
 			if err != nil || size <= 0 || size > 8 {
 				return fmt.Errorf("size must be a positive integer (<=8)")
 			}
+		case "-x":
+			isExpr = true
+			break loop // remaining args are going to be interpreted as expression
 		default:
-			if i != len(v)-1 {
-				return fmt.Errorf("unknown option %q", v[i])
-			}
-			// TODO, maybe we can support expression.
-			address, err = strconv.ParseUint(v[len(v)-1], 0, 64)
-			if err != nil {
-				return fmt.Errorf("convert address into uintptr type failed, %s", err)
+			if len(args) > 0 {
+				return fmt.Errorf("unknown option %q", args[0])
 			}
+			args = []string{cmd}
+			break loop // only one arg left to be evaluated as a uint
 		}
 	}
 
@@ -1623,10 +1753,39 @@ func examineMemoryCmd(t *Term, ctx callContext, args string) error {
 		return fmt.Errorf("read memory range (count*size) must be less than or equal to 1000 bytes")
 	}
 
-	if address == 0 {
+	if len(args) == 0 {
 		return fmt.Errorf("no address specified")
 	}
 
+	if isExpr {
+		expr := strings.Join(args, " ")
+		val, err := t.client.EvalVariable(ctx.Scope, expr, t.loadConfig())
+		if err != nil {
+			return err
+		}
+
+		// "-x &myVar" or "-x myPtrVar"
+		if val.Kind == reflect.Ptr {
+			if len(val.Children) < 1 {
+				return fmt.Errorf("bug? invalid pointer: %#v", val)
+			}
+			address = val.Children[0].Addr
+			// "-x 0xc000079f20 + 8" or -x 824634220320 + 8
+		} else if val.Kind == reflect.Int && val.Value != "" {
+			address, err = strconv.ParseUint(val.Value, 0, 64)
+			if err != nil {
+				return fmt.Errorf("bad expression result: %q: %s", val.Value, err)
+			}
+		} else {
+			return fmt.Errorf("unsupported expression type: %s", val.Kind)
+		}
+	} else {
+		address, err = strconv.ParseUint(args[0], 0, 64)
+		if err != nil {
+			return fmt.Errorf("convert address into uintptr type failed, %s", err)
+		}
+	}
+
 	memArea, isLittleEndian, err := t.client.ExamineMemory(address, count*size)
 	if err != nil {
 		return err
@@ -1635,6 +1794,17 @@ func examineMemoryCmd(t *Term, ctx callContext, args string) error {
 	return nil
 }
 
+func parseFormatArg(args string) (fmtstr, argsOut string) {
+	if len(args) < 1 || args[0] != '%' {
+		return "", args
+	}
+	v := strings.SplitN(args, " ", 2)
+	if len(v) == 1 {
+		return v[0], ""
+	}
+	return v[0], v[1]
+}
+
 func printVar(t *Term, ctx callContext, args string) error {
 	if len(args) == 0 {
 		return fmt.Errorf("not enough arguments")
@@ -1643,12 +1813,13 @@ func printVar(t *Term, ctx callContext, args string) error {
 		ctx.Breakpoint.Variables = append(ctx.Breakpoint.Variables, args)
 		return nil
 	}
+	fmtstr, args := parseFormatArg(args)
 	val, err := t.client.EvalVariable(ctx.Scope, args, t.loadConfig())
 	if err != nil {
 		return err
 	}
 
-	fmt.Println(val.MultilineString(""))
+	fmt.Println(val.MultilineString("", fmtstr))
 	return nil
 }
 
@@ -1660,6 +1831,10 @@ func whatisCommand(t *Term, ctx callContext, args string) error {
 	if err != nil {
 		return err
 	}
+	if val.Flags&api.VariableCPURegister != 0 {
+		fmt.Println("CPU Register")
+		return nil
+	}
 	if val.Type != "" {
 		fmt.Println(val.Type)
 	}
@@ -1708,7 +1883,7 @@ func printFilteredVariables(varType string, vars []api.Variable, filter string,
 			if cfg == ShortLoadConfig {
 				fmt.Printf("%s = %s\n", name, v.SinglelineString())
 			} else {
-				fmt.Printf("%s = %s\n", name, v.MultilineString(""))
+				fmt.Printf("%s = %s\n", name, v.MultilineString("", ""))
 			}
 		}
 	}
@@ -2104,6 +2279,10 @@ func digits(n int) int {
 const stacktraceTruncatedMessage = "(truncated)"
 
 func printStack(t *Term, out io.Writer, stack []api.Stackframe, ind string, offsets bool) {
+	PrintStack(t.formatPath, out, stack, ind, offsets)
+}
+
+func PrintStack(formatPath func(string) string, out io.Writer, stack []api.Stackframe, ind string, offsets bool) {
 	if len(stack) == 0 {
 		return
 	}
@@ -2126,7 +2305,7 @@ func printStack(t *Term, out io.Writer, stack []api.Stackframe, ind string, offs
 			continue
 		}
 		fmt.Fprintf(out, fmtstr, ind, i, stack[i].PC, stack[i].Function.Name())
-		fmt.Fprintf(out, "%sat %s:%d\n", s, t.formatPath(stack[i].File), stack[i].Line)
+		fmt.Fprintf(out, "%sat %s:%d\n", s, formatPath(stack[i].File), stack[i].Line)
 
 		if offsets {
 			fmt.Fprintf(out, "%sframe: %+#x frame pointer %+#x\n", s, stack[i].FrameOffset, stack[i].FramePointerOffset)
@@ -2140,8 +2319,8 @@ func printStack(t *Term, out io.Writer, stack []api.Stackframe, ind string, offs
 				continue
 			}
 			fmt.Fprintf(out, "%s%#016x in %s\n", deferHeader, d.DeferredLoc.PC, d.DeferredLoc.Function.Name())
-			fmt.Fprintf(out, "%sat %s:%d\n", s2, t.formatPath(d.DeferredLoc.File), d.DeferredLoc.Line)
-			fmt.Fprintf(out, "%sdeferred by %s at %s:%d\n", s2, d.DeferLoc.Function.Name(), t.formatPath(d.DeferLoc.File), d.DeferLoc.Line)
+			fmt.Fprintf(out, "%sat %s:%d\n", s2, formatPath(d.DeferredLoc.File), d.DeferredLoc.Line)
+			fmt.Fprintf(out, "%sdeferred by %s at %s:%d\n", s2, d.DeferLoc.Function.Name(), formatPath(d.DeferLoc.File), d.DeferLoc.Line)
 		}
 
 		for j := range stack[i].Arguments {
@@ -2194,7 +2373,7 @@ func printcontext(t *Term, state *api.DebuggerState) {
 
 	if th.File == "" {
 		fmt.Printf("Stopped at: 0x%x\n", state.CurrentThread.PC)
-		t.Println("=>", "no source available")
+		_ = colorize.Print(t.stdout, "", bytes.NewReader([]byte("no source available")), 1, 10, 1, nil)
 		return
 	}
 
@@ -2218,7 +2397,7 @@ func printReturnValues(th *api.Thread) {
 	}
 	fmt.Println("Values returned:")
 	for _, v := range th.ReturnValues {
-		fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t"))
+		fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t", ""))
 	}
 	fmt.Println()
 }
@@ -2318,13 +2497,13 @@ func printBreakpointInfo(t *Term, th *api.Thread, tracepointOnNewline bool) {
 
 	for _, v := range bpi.Variables {
 		tracepointnl()
-		fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t"))
+		fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t", ""))
 	}
 
 	for _, v := range bpi.Locals {
 		tracepointnl()
 		if *bp.LoadLocals == longLoadConfig {
-			fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t"))
+			fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t", ""))
 		} else {
 			fmt.Printf("\t%s: %s\n", v.Name, v.SinglelineString())
 		}
@@ -2333,7 +2512,7 @@ func printBreakpointInfo(t *Term, th *api.Thread, tracepointOnNewline bool) {
 	if bp.LoadArgs != nil && *bp.LoadArgs == longLoadConfig {
 		for _, v := range bpi.Arguments {
 			tracepointnl()
-			fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t"))
+			fmt.Printf("\t%s: %s\n", v.Name, v.MultilineString("\t", ""))
 		}
 	}
 
@@ -2371,6 +2550,13 @@ func printfile(t *Term, filename string, line int, showArrow bool) error {
 	if filename == "" {
 		return nil
 	}
+
+	lineCount := t.conf.GetSourceListLineCount()
+	arrowLine := 0
+	if showArrow {
+		arrowLine = line
+	}
+
 	file, err := os.Open(t.substitutePath(filename))
 	if err != nil {
 		return err
@@ -2383,38 +2569,7 @@ func printfile(t *Term, filename string, line int, showArrow bool) error {
 		fmt.Println("Warning: listing may not match stale executable")
 	}
 
-	lineCount := t.conf.GetSourceListLineCount()
-
-	buf := bufio.NewScanner(file)
-	l := line
-	for i := 1; i < l-lineCount; i++ {
-		if !buf.Scan() {
-			return nil
-		}
-	}
-
-	s := l - lineCount
-	if s < 1 {
-		s = 1
-	}
-
-	for i := s; i <= l+lineCount; i++ {
-		if !buf.Scan() {
-			return nil
-		}
-
-		var prefix string
-		if showArrow {
-			prefix = "  "
-			if i == l {
-				prefix = "=>"
-			}
-		}
-
-		prefix = fmt.Sprintf("%s%4d:\t", prefix, i)
-		t.Println(prefix, buf.Text())
-	}
-	return nil
+	return colorize.Print(t.stdout, file.Name(), file, line-lineCount, line+lineCount+1, arrowLine, t.colorEscapes)
 }
 
 // ExitRequestError is returned when the user
@@ -2583,10 +2738,11 @@ func display(t *Term, ctx callContext, args string) error {
 
 	case strings.HasPrefix(args, addOption):
 		args = strings.TrimSpace(args[len(addOption):])
+		fmtstr, args := parseFormatArg(args)
 		if args == "" {
 			return fmt.Errorf("not enough arguments")
 		}
-		t.addDisplay(args)
+		t.addDisplay(args, fmtstr)
 		t.printDisplay(len(t.displays) - 1)
 
 	case strings.HasPrefix(args, delOption):
@@ -2603,6 +2759,36 @@ func display(t *Term, ctx callContext, args string) error {
 	return nil
 }
 
+func dump(t *Term, ctx callContext, args string) error {
+	if args == "" {
+		return fmt.Errorf("not enough arguments")
+	}
+	dumpState, err := t.client.CoreDumpStart(args)
+	if err != nil {
+		return err
+	}
+	for {
+		if dumpState.ThreadsDone != dumpState.ThreadsTotal {
+			fmt.Printf("\rDumping threads %d / %d...", dumpState.ThreadsDone, dumpState.ThreadsTotal)
+		} else {
+			fmt.Printf("\rDumping memory %d / %d...", dumpState.MemDone, dumpState.MemTotal)
+		}
+		if !dumpState.Dumping {
+			break
+		}
+		dumpState = t.client.CoreDumpWait(1000)
+	}
+	fmt.Printf("\n")
+	if dumpState.Err != "" {
+		fmt.Printf("error dumping: %s\n", dumpState.Err)
+	} else if !dumpState.AllDone {
+		fmt.Printf("canceled\n")
+	} else if dumpState.MemDone != dumpState.MemTotal {
+		fmt.Printf("Core dump could be incomplete\n")
+	}
+	return nil
+}
+
 func formatBreakpointName(bp *api.Breakpoint, upcase bool) string {
 	thing := "breakpoint"
 	if bp.Tracepoint {
@@ -2615,7 +2801,11 @@ func formatBreakpointName(bp *api.Breakpoint, upcase bool) string {
 	if id == "" {
 		id = strconv.Itoa(bp.ID)
 	}
-	return fmt.Sprintf("%s %s", thing, id)
+	state := "(enabled)"
+	if bp.Disabled {
+		state = "(disabled)"
+	}
+	return fmt.Sprintf("%s %s %s", thing, id, state)
 }
 
 func (t *Term) formatBreakpointLocation(bp *api.Breakpoint) string {
