@@ -37,6 +37,12 @@ var (
 	// ErrNotRecording is returned when StopRecording is called while the
 	// debugger is not recording the target.
 	ErrNotRecording = errors.New("debugger is not recording")
+
+	// ErrCoreDumpInProgress is returned when a core dump is already in progress.
+	ErrCoreDumpInProgress = errors.New("core dump in progress")
+
+	// ErrCoreDumpNotSupported is returned when core dumping is not supported
+	ErrCoreDumpNotSupported = errors.New("core dumping not supported")
 )
 
 // Debugger service.
@@ -62,6 +68,12 @@ type Debugger struct {
 
 	stopRecording func() error
 	recordMutex   sync.Mutex
+
+	dumpState proc.DumpState
+	// Debugger keeps a map of disabled breakpoints
+	// so lower layers like proc doesn't need to deal
+	// with them
+	disabledBreakpoints map[int]*api.Breakpoint
 }
 
 type ExecuteKind int
@@ -190,6 +202,9 @@ func New(config *Config, processArgs []string) (*Debugger, error) {
 			return nil, err
 		}
 	}
+
+	d.disabledBreakpoints = make(map[int]*api.Breakpoint)
+
 	return d, nil
 }
 
@@ -361,8 +376,6 @@ func (d *Debugger) LastModified() time.Time {
 	return d.target.BinInfo().LastModified()
 }
 
-const deferReturn = "runtime.deferreturn"
-
 // FunctionReturnLocations returns all return locations
 // for the given function, a list of addresses corresponding
 // to 'ret' or 'call runtime.deferreturn'.
@@ -405,9 +418,12 @@ func (d *Debugger) FunctionReturnLocations(fnName string) ([]uint64, error) {
 // If `kill` is true we will kill the process after
 // detaching.
 func (d *Debugger) Detach(kill bool) error {
+	d.log.Debug("detaching")
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
-
+	if ok, _ := d.target.Valid(); !ok {
+		return nil
+	}
 	return d.detach(kill)
 }
 
@@ -491,23 +507,29 @@ func (d *Debugger) Restart(rerecord bool, pos string, resetArgs bool, newArgs []
 	}
 
 	discarded := []api.DiscardedBreakpoint{}
-	for _, oldBp := range api.ConvertBreakpoints(d.breakpoints()) {
+	breakpoints := api.ConvertBreakpoints(d.breakpoints())
+	d.target = p
+	maxID := 0
+	for _, oldBp := range breakpoints {
 		if oldBp.ID < 0 {
 			continue
 		}
+		if oldBp.ID > maxID {
+			maxID = oldBp.ID
+		}
 		if len(oldBp.File) > 0 {
 			addrs, err := proc.FindFileLocation(p, oldBp.File, oldBp.Line)
 			if err != nil {
 				discarded = append(discarded, api.DiscardedBreakpoint{Breakpoint: oldBp, Reason: err.Error()})
 				continue
 			}
-			createLogicalBreakpoint(p, addrs, oldBp)
+			createLogicalBreakpoint(d, addrs, oldBp, oldBp.ID)
 		} else {
 			// Avoid setting a breakpoint based on address when rebuilding
 			if rebuild {
 				continue
 			}
-			newBp, err := p.SetBreakpoint(oldBp.Addr, proc.UserBreakpoint, nil)
+			newBp, err := p.SetBreakpointWithID(oldBp.ID, oldBp.Addr)
 			if err != nil {
 				return nil, err
 			}
@@ -516,13 +538,18 @@ func (d *Debugger) Restart(rerecord bool, pos string, resetArgs bool, newArgs []
 			}
 		}
 	}
-	d.target = p
+	for _, bp := range d.disabledBreakpoints {
+		if bp.ID > maxID {
+			maxID = bp.ID
+		}
+	}
+	d.target.SetNextBreakpointID(maxID)
 	return discarded, nil
 }
 
 // State returns the current state of the debugger.
 func (d *Debugger) State(nowait bool) (*api.DebuggerState, error) {
-	if d.isRunning() && nowait {
+	if d.IsRunning() && nowait {
 		return &api.DebuggerState{Running: true}, nil
 	}
 
@@ -530,6 +557,12 @@ func (d *Debugger) State(nowait bool) (*api.DebuggerState, error) {
 		return &api.DebuggerState{Recording: true}, nil
 	}
 
+	d.dumpState.Mutex.Lock()
+	if d.dumpState.Dumping && nowait {
+		return &api.DebuggerState{CoreDumping: true}, nil
+	}
+	d.dumpState.Mutex.Unlock()
+
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
 	return d.state(nil)
@@ -551,7 +584,7 @@ func (d *Debugger) state(retLoadCfg *proc.LoadConfig) (*api.DebuggerState, error
 
 	exited := false
 	if _, err := d.target.Valid(); err != nil {
-		_, exited = err.(*proc.ErrProcessExited)
+		_, exited = err.(proc.ErrProcessExited)
 	}
 
 	state = &api.DebuggerState{
@@ -562,6 +595,7 @@ func (d *Debugger) state(retLoadCfg *proc.LoadConfig) (*api.DebuggerState, error
 	for _, thread := range d.target.ThreadList() {
 		th := api.ConvertThread(thread)
 
+		th.CallReturn = thread.Common().CallReturn
 		if retLoadCfg != nil {
 			th.ReturnValues = api.ConvertVars(thread.Common().ReturnValues(*retLoadCfg))
 		}
@@ -595,7 +629,7 @@ func (d *Debugger) CreateBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoin
 		if err = api.ValidBreakpointName(requestedBp.Name); err != nil {
 			return nil, err
 		}
-		if d.findBreakpointByName(requestedBp.Name) != nil {
+		if (d.findBreakpointByName(requestedBp.Name) != nil) || (d.findDisabledBreakpointByName(requestedBp.Name) != nil) {
 			return nil, errors.New("breakpoint name already exists")
 		}
 	}
@@ -628,7 +662,7 @@ func (d *Debugger) CreateBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoin
 		return nil, err
 	}
 
-	createdBp, err := createLogicalBreakpoint(d.target, addrs, requestedBp)
+	createdBp, err := createLogicalBreakpoint(d, addrs, requestedBp, 0)
 	if err != nil {
 		return nil, err
 	}
@@ -638,11 +672,21 @@ func (d *Debugger) CreateBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoin
 
 // createLogicalBreakpoint creates one physical breakpoint for each address
 // in addrs and associates all of them with the same logical breakpoint.
-func createLogicalBreakpoint(p *proc.Target, addrs []uint64, requestedBp *api.Breakpoint) (*api.Breakpoint, error) {
+func createLogicalBreakpoint(d *Debugger, addrs []uint64, requestedBp *api.Breakpoint, id int) (*api.Breakpoint, error) {
+	p := d.target
+
+	if dbp, ok := d.disabledBreakpoints[requestedBp.ID]; ok {
+		return dbp, proc.BreakpointExistsError{File: dbp.File, Line: dbp.Line, Addr: dbp.Addr}
+	}
+
 	bps := make([]*proc.Breakpoint, len(addrs))
 	var err error
 	for i := range addrs {
-		bps[i], err = p.SetBreakpoint(addrs[i], proc.UserBreakpoint, nil)
+		if id > 0 {
+			bps[i], err = p.SetBreakpointWithID(id, addrs[i])
+		} else {
+			bps[i], err = p.SetBreakpoint(addrs[i], proc.UserBreakpoint, nil)
+		}
 		if err != nil {
 			break
 		}
@@ -669,6 +713,7 @@ func createLogicalBreakpoint(p *proc.Target, addrs []uint64, requestedBp *api.Br
 		}
 		return nil, err
 	}
+
 	createdBp := api.ConvertBreakpoints(bps)
 	return createdBp[0], nil // we created a single logical breakpoint, the slice here will always have len == 1
 }
@@ -679,22 +724,39 @@ func isBreakpointExistsErr(err error) bool {
 }
 
 // AmendBreakpoint will update the breakpoint with the matching ID.
+// It also enables or disables the breakpoint.
 func (d *Debugger) AmendBreakpoint(amend *api.Breakpoint) error {
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
 
 	originals := d.findBreakpoint(amend.ID)
-	if originals == nil {
+	_, disabled := d.disabledBreakpoints[amend.ID]
+	if originals == nil && !disabled {
 		return fmt.Errorf("no breakpoint with ID %d", amend.ID)
 	}
 	if err := api.ValidBreakpointName(amend.Name); err != nil {
 		return err
 	}
+	if !amend.Disabled && disabled { // enable the breakpoint
+		bp, err := d.target.SetBreakpointWithID(amend.ID, amend.Addr)
+		if err != nil {
+			return err
+		}
+		copyBreakpointInfo(bp, amend)
+		delete(d.disabledBreakpoints, amend.ID)
+	}
+	if amend.Disabled && !disabled { // disable the breakpoint
+		if _, err := d.clearBreakpoint(amend); err != nil {
+			return err
+		}
+		d.disabledBreakpoints[amend.ID] = amend
+	}
 	for _, original := range originals {
 		if err := copyBreakpointInfo(original, amend); err != nil {
 			return err
 		}
 	}
+
 	return nil
 }
 
@@ -726,6 +788,15 @@ func copyBreakpointInfo(bp *proc.Breakpoint, requested *api.Breakpoint) (err err
 func (d *Debugger) ClearBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoint, error) {
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
+	return d.clearBreakpoint(requestedBp)
+}
+
+// clearBreakpoint clears a breakpoint, we can consume this function to avoid locking a goroutine
+func (d *Debugger) clearBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoint, error) {
+	if bp, ok := d.disabledBreakpoints[requestedBp.ID]; ok {
+		delete(d.disabledBreakpoints, bp.ID)
+		return bp, nil
+	}
 
 	var bps []*proc.Breakpoint
 	var errs []error
@@ -778,7 +849,14 @@ func (d *Debugger) ClearBreakpoint(requestedBp *api.Breakpoint) (*api.Breakpoint
 func (d *Debugger) Breakpoints() []*api.Breakpoint {
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
-	return api.ConvertBreakpoints(d.breakpoints())
+
+	bps := api.ConvertBreakpoints(d.breakpoints())
+
+	for _, bp := range d.disabledBreakpoints {
+		bps = append(bps, bp)
+	}
+
+	return bps
 }
 
 func (d *Debugger) breakpoints() []*proc.Breakpoint {
@@ -797,6 +875,7 @@ func (d *Debugger) FindBreakpoint(id int) *api.Breakpoint {
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
 	bps := api.ConvertBreakpoints(d.findBreakpoint(id))
+	bps = append(bps, d.findDisabledBreakpoint(id)...)
 	if len(bps) <= 0 {
 		return nil
 	}
@@ -806,18 +885,33 @@ func (d *Debugger) FindBreakpoint(id int) *api.Breakpoint {
 func (d *Debugger) findBreakpoint(id int) []*proc.Breakpoint {
 	var bps []*proc.Breakpoint
 	for _, bp := range d.target.Breakpoints().M {
-		if bp.LogicalID == id {
+		if bp.IsUser() && bp.LogicalID == id {
 			bps = append(bps, bp)
 		}
 	}
 	return bps
 }
 
+func (d *Debugger) findDisabledBreakpoint(id int) []*api.Breakpoint {
+	var bps []*api.Breakpoint
+	for _, dbp := range d.disabledBreakpoints {
+		if dbp.ID == id {
+			bps = append(bps, dbp)
+		}
+	}
+	return bps
+}
+
 // FindBreakpointByName returns the breakpoint specified by 'name'
 func (d *Debugger) FindBreakpointByName(name string) *api.Breakpoint {
 	d.targetMutex.Lock()
 	defer d.targetMutex.Unlock()
-	return d.findBreakpointByName(name)
+
+	bp := d.findBreakpointByName(name)
+	if bp == nil {
+		bp = d.findDisabledBreakpointByName(name)
+	}
+	return bp
 }
 
 func (d *Debugger) findBreakpointByName(name string) *api.Breakpoint {
@@ -835,6 +929,15 @@ func (d *Debugger) findBreakpointByName(name string) *api.Breakpoint {
 	return r[0] // there can only be one logical breakpoint with the same name
 }
 
+func (d *Debugger) findDisabledBreakpointByName(name string) *api.Breakpoint {
+	for _, dbp := range d.disabledBreakpoints {
+		if dbp.Name == name {
+			return dbp
+		}
+	}
+	return nil
+}
+
 // Threads returns the threads of the target process.
 func (d *Debugger) Threads() ([]proc.Thread, error) {
 	d.targetMutex.Lock()
@@ -864,20 +967,28 @@ func (d *Debugger) FindThread(id int) (proc.Thread, error) {
 	return nil, nil
 }
 
+// FindGoroutine returns the goroutine for the given 'id'.
+func (d *Debugger) FindGoroutine(id int) (*proc.G, error) {
+	d.targetMutex.Lock()
+	defer d.targetMutex.Unlock()
+
+	return proc.FindGoroutine(d.target, id)
+}
+
 func (d *Debugger) setRunning(running bool) {
 	d.runningMutex.Lock()
 	d.running = running
 	d.runningMutex.Unlock()
 }
 
-func (d *Debugger) isRunning() bool {
+func (d *Debugger) IsRunning() bool {
 	d.runningMutex.Lock()
 	defer d.runningMutex.Unlock()
 	return d.running
 }
 
 // Command handles commands which control the debugger lifecycle
-func (d *Debugger) Command(command *api.DebuggerCommand) (*api.DebuggerState, error) {
+func (d *Debugger) Command(command *api.DebuggerCommand, resumeNotify chan struct{}) (*api.DebuggerState, error) {
 	var err error
 
 	if command.Name == api.Halt {
@@ -900,6 +1011,12 @@ func (d *Debugger) Command(command *api.DebuggerCommand) (*api.DebuggerState, er
 	d.setRunning(true)
 	defer d.setRunning(false)
 
+	if command.Name != api.SwitchGoroutine && command.Name != api.SwitchThread && command.Name != api.Halt {
+		d.target.ResumeNotify(resumeNotify)
+	} else if resumeNotify != nil {
+		close(resumeNotify)
+	}
+
 	switch command.Name {
 	case api.Continue:
 		d.log.Debug("continuing")
@@ -986,7 +1103,8 @@ func (d *Debugger) Command(command *api.DebuggerCommand) (*api.DebuggerState, er
 		withBreakpointInfo = false
 	case api.SwitchGoroutine:
 		d.log.Debugf("switching to goroutine %d", command.GoroutineID)
-		g, err := proc.FindGoroutine(d.target, command.GoroutineID)
+		var g *proc.G
+		g, err = proc.FindGoroutine(d.target, command.GoroutineID)
 		if err == nil {
 			err = d.target.SwitchGoroutine(g)
 		}
@@ -1203,8 +1321,7 @@ func (d *Debugger) ThreadRegisters(threadID int, floatingPoint bool) (*op.DwarfR
 	if err != nil {
 		return nil, err
 	}
-	dregs := d.target.BinInfo().Arch.RegistersToDwarfRegisters(0, regs)
-	return &dregs, nil
+	return d.target.BinInfo().Arch.RegistersToDwarfRegisters(0, regs), nil
 }
 
 // ScopeRegisters returns registers for the specified scope.
@@ -1248,6 +1365,18 @@ func (d *Debugger) FunctionArguments(goid, frame, deferredCall int, cfg proc.Loa
 	return s.FunctionArguments(cfg)
 }
 
+// Function returns the current function.
+func (d *Debugger) Function(goid, frame, deferredCall int, cfg proc.LoadConfig) (*proc.Function, error) {
+	d.targetMutex.Lock()
+	defer d.targetMutex.Unlock()
+
+	s, err := proc.ConvertEvalScope(d.target, goid, frame, deferredCall)
+	if err != nil {
+		return nil, err
+	}
+	return s.Fn, nil
+}
+
 // EvalVariableInScope will attempt to evaluate the variable represented by 'symbol'
 // in the scope provided.
 func (d *Debugger) EvalVariableInScope(goid, frame, deferredCall int, symbol string, cfg proc.LoadConfig) (*proc.Variable, error) {
@@ -1456,9 +1585,28 @@ func (d *Debugger) FindLocation(goid, frame, deferredCall int, locStr string, in
 		return nil, err
 	}
 
+	return d.findLocation(goid, frame, deferredCall, locStr, loc, includeNonExecutableLines, substitutePathRules)
+}
+
+// FindLocationSpec will find the location specified by 'locStr' and 'locSpec'.
+// 'locSpec' should be the result of calling 'locspec.Parse(locStr)'. 'locStr'
+// is also passed, because it made be used to broaden the search criteria, if
+// the parsed result did not find anything.
+func (d *Debugger) FindLocationSpec(goid, frame, deferredCall int, locStr string, locSpec locspec.LocationSpec, includeNonExecutableLines bool, substitutePathRules [][2]string) ([]api.Location, error) {
+	d.targetMutex.Lock()
+	defer d.targetMutex.Unlock()
+
+	if _, err := d.target.Valid(); err != nil {
+		return nil, err
+	}
+
+	return d.findLocation(goid, frame, deferredCall, locStr, locSpec, includeNonExecutableLines, substitutePathRules)
+}
+
+func (d *Debugger) findLocation(goid, frame, deferredCall int, locStr string, locSpec locspec.LocationSpec, includeNonExecutableLines bool, substitutePathRules [][2]string) ([]api.Location, error) {
 	s, _ := proc.ConvertEvalScope(d.target, goid, frame, deferredCall)
 
-	locs, err := loc.Find(d.target, d.processArgs, s, locStr, includeNonExecutableLines, substitutePathRules)
+	locs, err := locSpec.Find(d.target, d.processArgs, s, locStr, includeNonExecutableLines, substitutePathRules)
 	for i := range locs {
 		if locs[i].PC == 0 {
 			continue
@@ -1602,7 +1750,7 @@ func (d *Debugger) GetVersion(out *api.GetVersionOut) error {
 		}
 	}
 
-	if !d.isRecording() && !d.isRunning() {
+	if !d.isRecording() && !d.IsRunning() {
 		out.TargetGoVersion = d.target.BinInfo().Producer()
 	}
 
@@ -1650,6 +1798,77 @@ func (d *Debugger) UnlockTarget() {
 	d.targetMutex.Unlock()
 }
 
+// DumpStart starts a core dump to dest.
+func (d *Debugger) DumpStart(dest string) error {
+	d.targetMutex.Lock()
+	// targetMutex will only be unlocked when the dump is done
+
+	if !d.target.CanDump {
+		d.targetMutex.Unlock()
+		return ErrCoreDumpNotSupported
+	}
+
+	d.dumpState.Mutex.Lock()
+	defer d.dumpState.Mutex.Unlock()
+
+	if d.dumpState.Dumping {
+		d.targetMutex.Unlock()
+		return ErrCoreDumpInProgress
+	}
+
+	fh, err := os.Create(dest)
+	if err != nil {
+		d.targetMutex.Unlock()
+		return err
+	}
+
+	d.dumpState.Dumping = true
+	d.dumpState.AllDone = false
+	d.dumpState.Canceled = false
+	d.dumpState.DoneChan = make(chan struct{})
+	d.dumpState.ThreadsDone = 0
+	d.dumpState.ThreadsTotal = 0
+	d.dumpState.MemDone = 0
+	d.dumpState.MemTotal = 0
+	d.dumpState.Err = nil
+	go func() {
+		defer d.targetMutex.Unlock()
+		d.target.Dump(fh, 0, &d.dumpState)
+	}()
+
+	return nil
+}
+
+// DumpWait waits for the dump to finish, or for the duration of wait.
+// Returns the state of the dump.
+// If wait == 0 returns immediately.
+func (d *Debugger) DumpWait(wait time.Duration) *proc.DumpState {
+	d.dumpState.Mutex.Lock()
+	if !d.dumpState.Dumping {
+		d.dumpState.Mutex.Unlock()
+		return &d.dumpState
+	}
+	d.dumpState.Mutex.Unlock()
+
+	if wait > 0 {
+		alarm := time.After(wait)
+		select {
+		case <-alarm:
+		case <-d.dumpState.DoneChan:
+		}
+	}
+
+	return &d.dumpState
+}
+
+// DumpCancel canels a dump in progress
+func (d *Debugger) DumpCancel() error {
+	d.dumpState.Mutex.Lock()
+	d.dumpState.Canceled = true
+	d.dumpState.Mutex.Unlock()
+	return nil
+}
+
 func go11DecodeErrorCheck(err error) error {
 	if _, isdecodeerr := err.(dwarf.DecodeError); !isdecodeerr {
 		return err
