@@ -3,8 +3,11 @@ package rpc2
 import (
 	"errors"
 	"fmt"
+	"sort"
 	"time"
 
+	"github.com/go-delve/delve/pkg/dwarf/op"
+	"github.com/go-delve/delve/pkg/proc"
 	"github.com/go-delve/delve/service"
 	"github.com/go-delve/delve/service/api"
 	"github.com/go-delve/delve/service/debugger"
@@ -55,12 +58,7 @@ type DetachOut struct {
 
 // Detach detaches the debugger, optionally killing the process.
 func (s *RPCServer) Detach(arg DetachIn, out *DetachOut) error {
-	err := s.debugger.Detach(arg.Kill)
-	if s.config.DisconnectChan != nil {
-		close(s.config.DisconnectChan)
-		s.config.DisconnectChan = nil
-	}
-	return err
+	return s.debugger.Detach(arg.Kill)
 }
 
 type RestartIn struct {
@@ -68,7 +66,7 @@ type RestartIn struct {
 	// otherwise it's an event number. Only valid for recorded targets.
 	Position string
 
-	// ResetArgs tell whether NewArgs should take effect.
+	// ResetArgs tell whether NewArgs and NewRedirects should take effect.
 	ResetArgs bool
 	// NewArgs are arguments to launch a new process.  They replace only the
 	// argv[1] and later. Argv[0] cannot be changed.
@@ -79,6 +77,8 @@ type RestartIn struct {
 
 	// When Rebuild is set the process will be build again
 	Rebuild bool
+
+	NewRedirects [3]string
 }
 
 type RestartOut struct {
@@ -87,13 +87,14 @@ type RestartOut struct {
 
 // Restart restarts program.
 func (s *RPCServer) Restart(arg RestartIn, cb service.RPCCallback) {
+	close(cb.SetupDoneChan())
 	if s.config.Debugger.AttachPid != 0 {
 		cb.Return(nil, errors.New("cannot restart process Delve did not create"))
 		return
 	}
 	var out RestartOut
 	var err error
-	out.DiscardedBreakpoints, err = s.debugger.Restart(arg.Rerecord, arg.Position, arg.ResetArgs, arg.NewArgs, arg.Rebuild)
+	out.DiscardedBreakpoints, err = s.debugger.Restart(arg.Rerecord, arg.Position, arg.ResetArgs, arg.NewArgs, arg.NewRedirects, arg.Rebuild)
 	cb.Return(out, err)
 }
 
@@ -108,6 +109,7 @@ type StateOut struct {
 
 // State returns the current debugger state.
 func (s *RPCServer) State(arg StateIn, cb service.RPCCallback) {
+	close(cb.SetupDoneChan())
 	var out StateOut
 	st, err := s.debugger.State(arg.NonBlocking)
 	if err != nil {
@@ -124,7 +126,7 @@ type CommandOut struct {
 
 // Command interrupts, continues and steps through the program.
 func (s *RPCServer) Command(command api.DebuggerCommand, cb service.RPCCallback) {
-	st, err := s.debugger.Command(&command)
+	st, err := s.debugger.Command(&command, cb.SetupDoneChan())
 	if err != nil {
 		cb.Return(nil, err)
 		return
@@ -187,7 +189,11 @@ func (s *RPCServer) Stacktrace(arg StacktraceIn, out *StacktraceOut) error {
 		arg.Opts |= api.StacktraceReadDefers
 	}
 	var err error
-	out.Locations, err = s.debugger.Stacktrace(arg.Id, arg.Depth, arg.Opts, api.LoadConfigToProc(cfg))
+	rawlocs, err := s.debugger.Stacktrace(arg.Id, arg.Depth, arg.Opts)
+	if err != nil {
+		return err
+	}
+	out.Locations, err = s.debugger.ConvertStacktrace(rawlocs, api.LoadConfigToProc(cfg))
 	return err
 }
 
@@ -283,6 +289,38 @@ func (s *RPCServer) ClearBreakpoint(arg ClearBreakpointIn, out *ClearBreakpointO
 	return nil
 }
 
+type ToggleBreakpointIn struct {
+	Id   int
+	Name string
+}
+
+type ToggleBreakpointOut struct {
+	Breakpoint *api.Breakpoint
+}
+
+// ToggleBreakpoint toggles on or off a breakpoint by Name (if Name is not an
+// empty string) or by ID.
+func (s *RPCServer) ToggleBreakpoint(arg ToggleBreakpointIn, out *ToggleBreakpointOut) error {
+	var bp *api.Breakpoint
+	if arg.Name != "" {
+		bp = s.debugger.FindBreakpointByName(arg.Name)
+		if bp == nil {
+			return fmt.Errorf("no breakpoint with name %s", arg.Name)
+		}
+	} else {
+		bp = s.debugger.FindBreakpoint(arg.Id)
+		if bp == nil {
+			return fmt.Errorf("no breakpoint with id %d", arg.Id)
+		}
+	}
+	bp.Disabled = !bp.Disabled
+	if err := s.debugger.AmendBreakpoint(bp); err != nil {
+		return err
+	}
+	out.Breakpoint = bp
+	return nil
+}
+
 type AmendBreakpointIn struct {
 	Breakpoint api.Breakpoint
 }
@@ -318,8 +356,14 @@ type ListThreadsOut struct {
 
 // ListThreads lists all threads.
 func (s *RPCServer) ListThreads(arg ListThreadsIn, out *ListThreadsOut) (err error) {
-	out.Threads, err = s.debugger.Threads()
-	return err
+	threads, err := s.debugger.Threads()
+	if err != nil {
+		return err
+	}
+	s.debugger.LockTarget()
+	defer s.debugger.UnlockTarget()
+	out.Threads = api.ConvertThreads(threads)
+	return nil
 }
 
 type GetThreadIn struct {
@@ -339,7 +383,9 @@ func (s *RPCServer) GetThread(arg GetThreadIn, out *GetThreadOut) error {
 	if t == nil {
 		return fmt.Errorf("no thread with id %d", arg.Id)
 	}
-	out.Thread = t
+	s.debugger.LockTarget()
+	defer s.debugger.UnlockTarget()
+	out.Thread = api.ConvertThread(t)
 	return nil
 }
 
@@ -354,21 +400,11 @@ type ListPackageVarsOut struct {
 
 // ListPackageVars lists all package variables in the context of the current thread.
 func (s *RPCServer) ListPackageVars(arg ListPackageVarsIn, out *ListPackageVarsOut) error {
-	state, err := s.debugger.State(false)
+	vars, err := s.debugger.PackageVariables(arg.Filter, *api.LoadConfigToProc(&arg.Cfg))
 	if err != nil {
 		return err
 	}
-
-	current := state.CurrentThread
-	if current == nil {
-		return fmt.Errorf("no current thread")
-	}
-
-	vars, err := s.debugger.PackageVariables(current.ID, arg.Filter, *api.LoadConfigToProc(&arg.Cfg))
-	if err != nil {
-		return err
-	}
-	out.Variables = vars
+	out.Variables = api.ConvertVars(vars)
 	return nil
 }
 
@@ -395,11 +431,18 @@ func (s *RPCServer) ListRegisters(arg ListRegistersIn, out *ListRegistersOut) er
 		arg.ThreadID = state.CurrentThread.ID
 	}
 
-	regs, err := s.debugger.Registers(arg.ThreadID, arg.Scope, arg.IncludeFp)
+	var regs *op.DwarfRegisters
+	var err error
+
+	if arg.Scope != nil {
+		regs, err = s.debugger.ScopeRegisters(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, arg.IncludeFp)
+	} else {
+		regs, err = s.debugger.ThreadRegisters(arg.ThreadID, arg.IncludeFp)
+	}
 	if err != nil {
 		return err
 	}
-	out.Regs = regs
+	out.Regs = api.ConvertRegisters(regs, s.debugger.DwarfRegisterToString, arg.IncludeFp)
 	out.Registers = out.Regs.String()
 
 	return nil
@@ -416,11 +459,11 @@ type ListLocalVarsOut struct {
 
 // ListLocalVars lists all local variables in scope.
 func (s *RPCServer) ListLocalVars(arg ListLocalVarsIn, out *ListLocalVarsOut) error {
-	vars, err := s.debugger.LocalVariables(arg.Scope, *api.LoadConfigToProc(&arg.Cfg))
+	vars, err := s.debugger.LocalVariables(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, *api.LoadConfigToProc(&arg.Cfg))
 	if err != nil {
 		return err
 	}
-	out.Variables = vars
+	out.Variables = api.ConvertVars(vars)
 	return nil
 }
 
@@ -435,11 +478,11 @@ type ListFunctionArgsOut struct {
 
 // ListFunctionArgs lists all arguments to the current function
 func (s *RPCServer) ListFunctionArgs(arg ListFunctionArgsIn, out *ListFunctionArgsOut) error {
-	vars, err := s.debugger.FunctionArguments(arg.Scope, *api.LoadConfigToProc(&arg.Cfg))
+	vars, err := s.debugger.FunctionArguments(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, *api.LoadConfigToProc(&arg.Cfg))
 	if err != nil {
 		return err
 	}
-	out.Args = vars
+	out.Args = api.ConvertVars(vars)
 	return nil
 }
 
@@ -455,18 +498,18 @@ type EvalOut struct {
 
 // EvalVariable returns a variable in the specified context.
 //
-// See https://github.com/go-delve/delve/wiki/Expressions for
-// a description of acceptable values of arg.Expr.
+// See https://github.com/go-delve/delve/blob/master/Documentation/cli/expr.md
+// for a description of acceptable values of arg.Expr.
 func (s *RPCServer) Eval(arg EvalIn, out *EvalOut) error {
 	cfg := arg.Cfg
 	if cfg == nil {
 		cfg = &api.LoadConfig{FollowPointers: true, MaxVariableRecurse: 1, MaxStringLen: 64, MaxArrayValues: 64, MaxStructFields: -1}
 	}
-	v, err := s.debugger.EvalVariableInScope(arg.Scope, arg.Expr, *api.LoadConfigToProc(cfg))
+	v, err := s.debugger.EvalVariableInScope(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, arg.Expr, *api.LoadConfigToProc(cfg))
 	if err != nil {
 		return err
 	}
-	out.Variable = v
+	out.Variable = api.ConvertVar(v)
 	return nil
 }
 
@@ -482,7 +525,7 @@ type SetOut struct {
 // Set sets the value of a variable. Only numerical types and
 // pointers are currently supported.
 func (s *RPCServer) Set(arg SetIn, out *SetOut) error {
-	return s.debugger.SetVariableInScope(arg.Scope, arg.Symbol, arg.Value)
+	return s.debugger.SetVariableInScope(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, arg.Symbol, arg.Value)
 }
 
 type ListSourcesIn struct {
@@ -560,7 +603,9 @@ func (s *RPCServer) ListGoroutines(arg ListGoroutinesIn, out *ListGoroutinesOut)
 	if err != nil {
 		return err
 	}
-	out.Goroutines = gs
+	s.debugger.LockTarget()
+	defer s.debugger.UnlockTarget()
+	out.Goroutines = api.ConvertGoroutines(gs)
 	out.Nextg = nextg
 	return nil
 }
@@ -584,6 +629,13 @@ type FindLocationIn struct {
 	Scope                     api.EvalScope
 	Loc                       string
 	IncludeNonExecutableLines bool
+
+	// SubstitutePathRules is a slice of source code path substitution rules,
+	// the first entry of each pair is the path of a directory as it appears in
+	// the executable file (i.e. the location of a source file when the program
+	// was compiled), the second entry of each pair is the location of the same
+	// directory on the client system.
+	SubstitutePathRules [][2]string
 }
 
 type FindLocationOut struct {
@@ -605,7 +657,7 @@ type FindLocationOut struct {
 // NOTE: this function does not actually set breakpoints.
 func (c *RPCServer) FindLocation(arg FindLocationIn, out *FindLocationOut) error {
 	var err error
-	out.Locations, err = c.debugger.FindLocation(arg.Scope, arg.Loc, arg.IncludeNonExecutableLines)
+	out.Locations, err = c.debugger.FindLocation(arg.Scope.GoroutineID, arg.Scope.Frame, arg.Scope.DeferredCall, arg.Loc, arg.IncludeNonExecutableLines, arg.SubstitutePathRules)
 	return err
 }
 
@@ -628,8 +680,15 @@ type DisassembleOut struct {
 // Disassemble will also try to calculate the destination address of an absolute indirect CALL if it happens to be the instruction the selected goroutine is stopped at.
 func (c *RPCServer) Disassemble(arg DisassembleIn, out *DisassembleOut) error {
 	var err error
-	out.Disassemble, err = c.debugger.Disassemble(arg.Scope.GoroutineID, arg.StartPC, arg.EndPC, arg.Flavour)
-	return err
+	insts, err := c.debugger.Disassemble(arg.Scope.GoroutineID, arg.StartPC, arg.EndPC)
+	if err != nil {
+		return err
+	}
+	out.Disassemble = make(api.AsmInstructions, len(insts))
+	for i := range insts {
+		out.Disassemble[i] = api.ConvertAsmInstruction(insts[i], c.debugger.AsmInstructionText(&insts[i], proc.AssemblyFlavour(arg.Flavour)))
+	}
+	return nil
 }
 
 type RecordedIn struct {
@@ -668,8 +727,15 @@ type ListCheckpointsOut struct {
 
 func (s *RPCServer) ListCheckpoints(arg ListCheckpointsIn, out *ListCheckpointsOut) error {
 	var err error
-	out.Checkpoints, err = s.debugger.Checkpoints()
-	return err
+	cps, err := s.debugger.Checkpoints()
+	if err != nil {
+		return err
+	}
+	out.Checkpoints = make([]api.Checkpoint, len(cps))
+	for i := range cps {
+		out.Checkpoints[i] = api.Checkpoint(cps[i])
+	}
+	return nil
 }
 
 type ClearCheckpointIn struct {
@@ -738,7 +804,11 @@ type ListDynamicLibrariesOut struct {
 }
 
 func (s *RPCServer) ListDynamicLibraries(in ListDynamicLibrariesIn, out *ListDynamicLibrariesOut) error {
-	out.List = s.debugger.ListDynamicLibraries()
+	imgs := s.debugger.ListDynamicLibraries()
+	out.List = make([]api.Image, 0, len(imgs))
+	for i := range imgs {
+		out.List[i] = api.ConvertImage(imgs[i])
+	}
 	return nil
 }
 
@@ -758,19 +828,39 @@ type ListPackagesBuildInfoOut struct {
 // Note that the directory path is a best guess and may be wrong is a tool
 // other than cmd/go is used to perform the build.
 func (s *RPCServer) ListPackagesBuildInfo(in ListPackagesBuildInfoIn, out *ListPackagesBuildInfoOut) error {
-	out.List = s.debugger.ListPackagesBuildInfo(in.IncludeFiles)
+	pkgs := s.debugger.ListPackagesBuildInfo(in.IncludeFiles)
+	out.List = make([]api.PackageBuildInfo, 0, len(pkgs))
+	for _, pkg := range pkgs {
+		var files []string
+
+		if len(pkg.Files) > 0 {
+			files = make([]string, 0, len(pkg.Files))
+			for file := range pkg.Files {
+				files = append(files, file)
+			}
+		}
+
+		sort.Strings(files)
+
+		out.List = append(out.List, api.PackageBuildInfo{
+			ImportPath:    pkg.ImportPath,
+			DirectoryPath: pkg.DirectoryPath,
+			Files:         files,
+		})
+	}
 	return nil
 }
 
 // ExamineMemoryIn holds the arguments of ExamineMemory
 type ExamineMemoryIn struct {
-	Address uintptr
+	Address uint64
 	Length  int
 }
 
 // ExaminedMemoryOut holds the return values of ExamineMemory
 type ExaminedMemoryOut struct {
-	Mem []byte
+	Mem            []byte
+	IsLittleEndian bool
 }
 
 func (s *RPCServer) ExamineMemory(arg ExamineMemoryIn, out *ExaminedMemoryOut) error {
@@ -783,6 +873,8 @@ func (s *RPCServer) ExamineMemory(arg ExamineMemoryIn, out *ExaminedMemoryOut) e
 	}
 
 	out.Mem = Mem
+	out.IsLittleEndian = true //TODO: get byte order from debugger.target.BinInfo().Arch
+
 	return nil
 }
 
@@ -793,6 +885,7 @@ type StopRecordingOut struct {
 }
 
 func (s *RPCServer) StopRecording(arg StopRecordingIn, cb service.RPCCallback) {
+	close(cb.SetupDoneChan())
 	var out StopRecordingOut
 	err := s.debugger.StopRecording()
 	if err != nil {
@@ -801,3 +894,48 @@ func (s *RPCServer) StopRecording(arg StopRecordingIn, cb service.RPCCallback) {
 	}
 	cb.Return(out, nil)
 }
+
+type DumpStartIn struct {
+	Destination string
+}
+
+type DumpStartOut struct {
+	State api.DumpState
+}
+
+// DumpStart starts a core dump to arg.Destination.
+func (s *RPCServer) DumpStart(arg DumpStartIn, out *DumpStartOut) error {
+	err := s.debugger.DumpStart(arg.Destination)
+	if err != nil {
+		return err
+	}
+	out.State = *api.ConvertDumpState(s.debugger.DumpWait(0))
+	return nil
+}
+
+type DumpWaitIn struct {
+	Wait int
+}
+
+type DumpWaitOut struct {
+	State api.DumpState
+}
+
+// DumpWait waits for the core dump to finish or for arg.Wait milliseconds.
+// Wait == 0 means return immediately.
+// Returns the core dump status
+func (s *RPCServer) DumpWait(arg DumpWaitIn, out *DumpWaitOut) error {
+	out.State = *api.ConvertDumpState(s.debugger.DumpWait(time.Duration(arg.Wait) * time.Millisecond))
+	return nil
+}
+
+type DumpCancelIn struct {
+}
+
+type DumpCancelOut struct {
+}
+
+// DumpCancel cancels the core dump.
+func (s *RPCServer) DumpCancel(arg DumpCancelIn, out *DumpCancelOut) error {
+	return s.debugger.DumpCancel()
+}
